# Authors:
#   Alexander Bokovoy <abokovoy@redhat.com>
#   Tomas Babej <tbabej@redhat.com>
#
# Copyright (C) 2011-2014  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
This module contains default platform-specific implementations of system tasks.
'''

from __future__ import absolute_import

import os
import logging
import textwrap

from pkg_resources import parse_version

from ipaplatform.paths import paths
from ipapython import ipautil
from ipapython.ipachangeconf import IPAChangeConf

logger = logging.getLogger(__name__)


# TODO: Add other masters as FallbackDNS ?
RESOLVE1_IPA_CONF = textwrap.dedent("""
    # auto-generated by IPA installer
    [Resolve]
    # use local BIND instance
    DNS=127.0.0.1
    # make local BIND default DNS server, add search suffixes
    Domains=~. {searchdomains}
""")


class BaseTaskNamespace:

    def restore_context(self, filepath, force=False):
        """Restore SELinux security context on the given filepath.

        No return value expected.
        """
        raise NotImplementedError()

    def backup_hostname(self, fstore, statestore):
        """
        Backs up the current hostname in the statestore (so that it can be
        restored by the restore_hostname platform task).

        No return value expected.
        """

        raise NotImplementedError()

    def reload_systemwide_ca_store(self):
        """
        Reloads the systemwide CA store.

        Returns True if the operation succeeded, False otherwise.
        """

        raise NotImplementedError()

    def insert_ca_certs_into_systemwide_ca_store(self, ca_certs):
        """
        Adds CA certificates from 'ca_certs' to the systemwide CA store
        (if available on the platform).

        Returns True if the operation succeeded, False otherwise.
        """

        try:
            if self.platform_insert_ca_certs(ca_certs):
                return self.reload_systemwide_ca_store()
        except Exception:
            logger.exception('Could not populate systemwide CA store')

        return False

    def platform_insert_ca_certs(self, ca_certs):
        """
        Platform implementations override this method to implement
        population of the systemwide CA store.

        Returns True if changes were made to the CA store, False otherwise.

        Raises Exception if something went wrong.
        """
        raise NotImplementedError()

    def remove_ca_certs_from_systemwide_ca_store(self):
        """
        Removes IPA CA certificates from the systemwide CA store
        (if available on the platform).

        Returns True if the operation succeeded, False otherwise.
        """

        try:
            if self.platform_remove_ca_certs():
                return self.reload_systemwide_ca_store()
        except Exception:
            logger.exception(
                'Could not remove certificates from systemwide CA store'
            )

        return False

    def platform_remove_ca_certs(self):
        """
        Platform implementations override this method to implement
        removal of certificates from the systemwide CA store.

        Returns True if changes were made to the CA store, False otherwise.

        Raises Exception if something went wrong.
        """
        raise NotImplementedError()

    def get_svc_list_file(self):
        """
        Returns the path to the IPA service list file.
        """

        return paths.SVC_LIST_FILE

    def is_selinux_enabled(self):
        """Check if SELinux is available and enabled

        :return: True if SELinux is available and enabled
        """
        return False

    def check_selinux_status(self):
        """Checks if SELinux is available on the platform.

        If it is, this task also makes sure that restorecon tool is available.

        If SELinux is available, but restorcon tool is not installed, raises
        an RuntimeError, which suggest installing the package containing
        restorecon and rerunning the installation.

        :return: True if SELinux is available and enabled
        """
        raise NotImplementedError()

    def check_ipv6_stack_enabled(self):
        """Check whether IPv6 kernel module is loaded"""

        raise NotImplementedError()

    def detect_container(self):
        """Check if running inside a container

        :returns: container runtime or None
        :rtype: str, None
        """
        raise NotImplementedError

    def restore_hostname(self, fstore, statestore):
        """
        Restores the original hostname as backed up in the
        backup_hostname platform task.
        """

        raise NotImplementedError()

    def restore_pre_ipa_client_configuration(self, fstore, statestore,
                                             was_sssd_installed,
                                             was_sssd_configured):
        """
        Restores the pre-ipa-client configuration that was modified by the
        following platform tasks:
            modify_nsswitch_pam_stack
            modify_pam_to_use_krb5
        """

        raise NotImplementedError()

    def set_nisdomain(self, nisdomain):
        """
        Sets the NIS domain name to 'nisdomain'.
        """

        raise NotImplementedError()

    def modify_nsswitch_pam_stack(self, sssd, mkhomedir, statestore,
                                  sudo=True):
        """
        If sssd flag is true, configure pam and nsswitch so that SSSD is used
        for retrieving user information and authentication.

        Otherwise, configure pam and nsswitch to leverage pure LDAP.
        """

        raise NotImplementedError()

    def modify_pam_to_use_krb5(self, statestore):
        """
        Configure pam stack to allow kerberos authentication.
        """

        raise NotImplementedError()

    def is_nosssd_supported(self):
        """
        Check if the flag --no-sssd is supported for client install.
        """
        return True

    def is_mkhomedir_supported(self):
        """
        Check if the flag --mkhomedir is supported for client install.
        """
        return True

    def backup_auth_configuration(self, path):
        """
        Create backup of access control configuration.
        :param path: store the backup here. This will be passed to
        restore_auth_configuration as well.
        """
        raise NotImplementedError()

    def restore_auth_configuration(self, path):
        """
        Restore backup of access control configuration.
        :param path: restore the backup from here.
        """
        raise NotImplementedError()

    def migrate_auth_configuration(self, statestore):
        """
        Migrate pam stack configuration to authselect.
        """

    def set_selinux_booleans(self, required_settings, backup_func=None):
        """Set the specified SELinux booleans

        :param required_settings: A dictionary mapping the boolean names
                                  to desired_values.
                                  The desired value can be 'on' or 'off',
                                  or None to leave the setting unchanged.

        :param backup_func: A function called for each boolean with two
                            arguments: the name and the previous value

        If SELinux is disabled, return False; on success returns True.

        If setting the booleans fails,
        an ipapython.errors.SetseboolError is raised.
        """

        raise NotImplementedError()

    @staticmethod
    def parse_ipa_version(version):
        """
        :param version: textual version
        :return: object implementing proper __cmp__ method for version compare
        """
        return parse_version(version)

    def set_hostname(self, hostname):
        """
        Set hostname for the system

        No return value expected, raise CalledProcessError when error occurred
        """
        raise NotImplementedError()

    def configure_httpd_service_ipa_conf(self):
        """Configure httpd service to work with IPA"""
        raise NotImplementedError()

    def configure_http_gssproxy_conf(self, ipauser):
        raise NotImplementedError()

    def remove_httpd_service_ipa_conf(self):
        """Remove configuration of httpd service of IPA"""
        raise NotImplementedError()

    def configure_httpd_wsgi_conf(self):
        """Configure WSGI for correct Python version"""
        raise NotImplementedError()

    def configure_httpd_protocol(self):
        """Configure TLS protocols in Apache"""
        raise NotImplementedError()

    def is_fips_enabled(self):
        return False

    def add_user_to_group(self, user, group):
        logger.debug('Adding user %s to group %s', user, group)
        args = [paths.USERMOD, '-a', '-G', group, user]
        try:
            ipautil.run(args)
            logger.debug('Done adding user to group')
        except ipautil.CalledProcessError as e:
            logger.debug('Failed to add user to group: %s', e)

    def setup_httpd_logging(self):
        raise NotImplementedError()

    def systemd_daemon_reload(self):
        """Tell systemd to reload config files"""
        raise NotImplementedError

    def configure_dns_resolver(self, nameservers, searchdomains, *,
                               resolve1_enabled=False, fstore=None):
        """Configure global DNS resolver (e.g. /etc/resolv.conf)

        :param nameservers: list of IP addresses
        :param searchdomains: list of search domaons
        :param resolve1_enabled: is systemd-resolved enabled?
        :param fstore: optional file store for backup
        """
        if resolve1_enabled:
            # break circular import
            from ipaplatform.services import knownservices

            confd = os.path.dirname(paths.SYSTEMD_RESOLVED_IPA_CONF)
            if not os.path.isdir(confd):
                os.mkdir(confd)
                # owned by root, readable by systemd-resolve user
                os.chmod(confd, 0o755)
                self.restore_context(confd, force=True)

            cfg = RESOLVE1_IPA_CONF.format(
                searchdomains=" ".join(searchdomains)
            )
            with open(paths.SYSTEMD_RESOLVED_IPA_CONF, "w") as f:
                os.fchmod(f.fileno(), 0o644)
                f.write(cfg)

            self.restore_context(
                paths.SYSTEMD_RESOLVED_IPA_CONF, force=True
            )

            knownservices["systemd-resolved"].reload_or_restart()

    def unconfigure_dns_resolver(self, fstore=None):
        """Unconfigure global DNS resolver (e.g. /etc/resolv.conf)

        :param fstore: optional file store for restore
        """
        if fstore is not None and fstore.has_file(paths.RESOLV_CONF):
            fstore.restore_file(paths.RESOLV_CONF)

        if os.path.isfile(paths.SYSTEMD_RESOLVED_IPA_CONF):
            # break circular import
            from ipaplatform.services import knownservices

            os.unlink(paths.SYSTEMD_RESOLVED_IPA_CONF)
            knownservices["systemd-resolved"].reload_or_restart()

    def configure_pkcs11_modules(self, fstore):
        """Disable p11-kit modules

        The p11-kit configuration injects p11-kit-proxy into all NSS
        databases. Amongst other p11-kit loads SoftHSM2 PKCS#11 provider.
        This interferes with 389-DS, certmonger, Dogtag and other services.
        For example certmonger tries to open OpenDNSSEC's SoftHSM2 token,
        although it doesn't use it at all. It also breaks Dogtag HSM support
        testing with SoftHSM2.

        IPA server does neither need nor use SoftHSM2 proxied by p11-kit.
        """
        raise NotImplementedError

    def restore_pkcs11_modules(self, fstore):
        """Restore global p11-kit modules for NSS
        """
        raise NotImplementedError

    def get_pkcs11_modules(self):
        """Return the list of module config files setup by IPA.
        """
        return ()

    def configure_nsswitch_database(self, fstore, database, services,
                                    preserve=True, append=True,
                                    default_value=()):
        """
        Edits the specified nsswitch.conf database (e.g. passwd, group,
        sudoers) to use the specified service(s).

        Arguments:
            fstore - FileStore to backup the nsswitch.conf
            database - database configuration that should be ammended,
                       e.g. 'sudoers'
            service - list of services that should be added, e.g. ['sss']
            preserve - if True, the already configured services will be
                       preserved

        The next arguments modify the behaviour if preserve=True:
            append - if True, the services will be appended, if False,
                     prepended
            default_value - list of services that are considered as default (if
                            the database is not mentioned in nsswitch.conf),
                            e.g. ['files']
        """

        # Backup the original version of nsswitch.conf, we're going to edit it
        # now
        if not fstore.has_file(paths.NSSWITCH_CONF):
            fstore.backup_file(paths.NSSWITCH_CONF)

        conf = IPAChangeConf("IPA Installer")
        conf.setOptionAssignment(':')

        if preserve:
            # Read the existing configuration
            with open(paths.NSSWITCH_CONF, 'r') as f:
                opts = conf.parse(f)
                raw_database_entry = conf.findOpts(opts, 'option', database)[1]

            # Detect the list of already configured services
            if not raw_database_entry:
                # If there is no database entry, database is not present in
                # the nsswitch.conf. Set the list of services to the
                # default list, if passed.
                configured_services = list(default_value)
            else:
                configured_services = raw_database_entry[
                    'value'].strip().split()

            # Make sure no service is added if already mentioned in the list
            added_services = [s for s in services
                              if s not in configured_services]

            # Prepend / append the list of new services
            if append:
                new_value = ' ' + ' '.join(configured_services +
                                           added_services)
            else:
                new_value = ' ' + ' '.join(added_services +
                                           configured_services)

        else:
            # Preserve not set, let's rewrite existing configuration
            new_value = ' ' + ' '.join(services)

        # Set new services as sources for database
        opts = [
            conf.setOption(database, new_value),
            conf.emptyLine(),
        ]

        conf.changeConf(paths.NSSWITCH_CONF, opts)
        logger.info("Configured %s in %s", database, paths.NSSWITCH_CONF)

    def enable_sssd_sudo(self, fstore):
        """Configure nsswitch.conf to use sssd for sudo"""
        self.configure_nsswitch_database(
            fstore, 'sudoers', ['sss'],
            default_value=['files'])

    def enable_ldap_automount(self, statestore):
        """
        Point automount to ldap in nsswitch.conf.
        This function is for non-SSSD setups only.
        """
        conf = IPAChangeConf("IPA Installer")
        conf.setOptionAssignment(':')

        with open(paths.NSSWITCH_CONF, 'r') as f:
            current_opts = conf.parse(f)
            current_nss_value = conf.findOpts(
                current_opts, name='automount', type='option'
            )[1]
            if current_nss_value is None:
                # no automount database present
                current_nss_value = False  # None cannot be backed up
            else:
                current_nss_value = current_nss_value['value']
            statestore.backup_state(
                'ipa-client-automount-nsswitch', 'previous-automount',
                current_nss_value
            )

        nss_value = ' files ldap'
        opts = [
            {
                'name': 'automount',
                'type': 'option',
                'action': 'set',
                'value': nss_value,
            },
            {'name': 'empty', 'type': 'empty'},
        ]
        conf.changeConf(paths.NSSWITCH_CONF, opts)

        logger.info("Configured %s", paths.NSSWITCH_CONF)

    def disable_ldap_automount(self, statestore):
        """Disable automount using LDAP"""
        if statestore.get_state(
            'ipa-client-automount-nsswitch', 'previous-automount'
        ) is False:
            # Previous nsswitch.conf had no automount database configured
            # so remove it.
            conf = IPAChangeConf("IPA automount installer")
            conf.setOptionAssignment(':')
            changes = [conf.rmOption('automount')]
            conf.changeConf(paths.NSSWITCH_CONF, changes)
            self.restore_context(paths.NSSWITCH_CONF)
            statestore.delete_state(
                'ipa-client-automount-nsswitch', 'previous-automount'
            )
        elif statestore.get_state(
            'ipa-client-automount-nsswitch', 'previous-automount'
        ) is not None:
            nss_value = statestore.get_state(
                'ipa-client-automount-nsswitch', 'previous-automount'
            )
            opts = [
                {
                    'name': 'automount',
                    'type': 'option',
                    'action': 'set',
                    'value': nss_value,
                },
                {'name': 'empty', 'type': 'empty'},
            ]
            conf = IPAChangeConf("IPA automount installer")
            conf.setOptionAssignment(':')
            conf.changeConf(paths.NSSWITCH_CONF, opts)
            self.restore_context(paths.NSSWITCH_CONF)
            statestore.delete_state(
                'ipa-client-automount-nsswitch', 'previous-automount'
            )

tasks = BaseTaskNamespace()



# Authors:
#   Petr Viktorin <pviktori@redhat.com>
#
# Copyright (C) 2013  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Host class for integration testing"""
import re
import subprocess
import tempfile

import ldap
import pytest_multihost.host

from ipaplatform.paths import paths
from ipapython import ipaldap

from .fips import (
    is_fips_enabled, enable_userspace_fips, disable_userspace_fips
)
from .transport import IPAOpenSSHTransport
from .resolver import resolver

FIPS_NOISE_RE = re.compile(br"FIPS mode initialized\r?\n?")


class LDAPClientWithoutCertCheck(ipaldap.LDAPClient):
    """Adds an option to disable certificate check for TLS connection

    To disable certificate validity check create client with added option
    no_certificate_check:
    client = LDAPClientWithoutCertCheck(..., no_certificate_check=True)
    """
    def __init__(self, *args, **kwargs):
        self._no_certificate_check = kwargs.pop(
            'no_certificate_check', False)
        super(LDAPClientWithoutCertCheck, self).__init__(*args, **kwargs)

    def _connect(self):
        if (self._start_tls and self.protocol == 'ldap' and
                self._no_certificate_check):
            with self.error_handler():
                conn = ipaldap.ldap_initialize(
                    self.ldap_uri, cacertfile=self._cacert)
                conn.set_option(ldap.OPT_X_TLS_REQUIRE_CERT,
                                ldap.OPT_X_TLS_NEVER)
                conn.set_option(ldap.OPT_X_TLS_NEWCTX, 0)
                conn.start_tls_s()
                return conn
        else:
            return super(LDAPClientWithoutCertCheck, self)._connect()


class Host(pytest_multihost.host.Host):
    """Representation of a remote IPA host"""

    transport_class = IPAOpenSSHTransport

    def __init__(self, domain, hostname, role, ip=None,
                 external_hostname=None, username=None, password=None,
                 test_dir=None, host_type=None):
        super().__init__(
            domain, hostname, role, ip=ip,
            external_hostname=external_hostname, username=username,
            password=password, test_dir=test_dir, host_type=host_type
        )
        self._fips_mode = None
        self._userspace_fips = False
        self.resolver = resolver(self)

    @property
    def is_fips_mode(self):
        """Check and cache if a system is in FIPS mode
        """
        if self._fips_mode is None:
            self._fips_mode = is_fips_enabled(self)
        return self._fips_mode

    @property
    def is_userspace_fips(self):
        """Check if host uses fake userspace FIPS
        """
        return self._userspace_fips

    def enable_userspace_fips(self):
        """Enable fake userspace FIPS mode

        The call has no effect if the system is already in FIPS mode.

        :return: True if system was modified, else None
        """
        if not self.is_fips_mode:
            enable_userspace_fips(self)
            self._fips_mode = True
            self._userspace_fips = True
            return True
        else:
            return False

    def disable_userspace_fips(self):
        """Disable fake userspace FIPS mode

        The call has no effect if userspace FIPS mode is not enabled.

        :return: True if system was modified, else None
        """
        if self.is_userspace_fips:
            disable_userspace_fips(self)
            self._userspace_fips = False
            self._fips_mode = False
            return True
        else:
            return False

    @staticmethod
    def _make_host(domain, hostname, role, ip, external_hostname):
        # We need to determine the type of the host, this depends on the domain
        # type, as we assume all Unix machines are in the Unix domain and
        # all Windows machine in a AD domain

        if domain.type == 'AD':
            cls = WinHost
        else:
            cls = Host

        return cls(
            domain,
            hostname,
            role,
            ip=ip,
            external_hostname=external_hostname
        )

    def ldap_connect(self):
        """Return an LDAPClient authenticated to this host as directory manager
        """
        self.log.info('Connecting to LDAP at %s', self.external_hostname)
        # get IPA CA cert to establish a secure connection
        cacert = self.get_file_contents(paths.IPA_CA_CRT)
        with tempfile.NamedTemporaryFile() as f:
            f.write(cacert)
            f.flush()

            hostnames_mismatch = self.hostname != self.external_hostname
            conn = LDAPClientWithoutCertCheck.from_hostname_secure(
                self.external_hostname,
                cacert=f.name,
                no_certificate_check=hostnames_mismatch)
            binddn = self.config.dirman_dn
            self.log.info('LDAP bind as %s', binddn)
            conn.simple_bind(binddn, self.config.dirman_password)

            # The CA cert file  has been loaded into the SSL_CTX and is no
            # longer required.

        return conn

    @classmethod
    def from_env(cls, env, domain, hostname, role, index, domain_index):
        from ipatests.pytest_ipa.integration.env_config import host_from_env
        return host_from_env(env, domain, hostname, role, index, domain_index)

    def to_env(self, **kwargs):
        from ipatests.pytest_ipa.integration.env_config import host_to_env
        return host_to_env(self, **kwargs)

    def run_command(self, argv, set_env=True, stdin_text=None,
                    log_stdout=True, raiseonerr=True,
                    cwd=None, bg=False, encoding='utf-8', ok_returncode=0):
        """Wrapper around run_command to log stderr on raiseonerr=True

        :param ok_returncode: return code considered to be correct,
                              you can pass an integer or sequence of integers
        """
        result = super().run_command(
            argv, set_env=set_env, stdin_text=stdin_text,
            log_stdout=log_stdout, raiseonerr=False, cwd=cwd, bg=bg,
            encoding=encoding
        )
        # in FIPS mode SSH may print noise to stderr, remove the string
        # "FIPS mode initialized" + optional newline.
        result.stderr_bytes = FIPS_NOISE_RE.sub(b'', result.stderr_bytes)
        try:
            result_ok = result.returncode in ok_returncode
        except TypeError:
            result_ok = result.returncode == ok_returncode
        if not result_ok and raiseonerr:
            result.log.error('stderr: %s', result.stderr_text)
            raise subprocess.CalledProcessError(
                result.returncode, argv,
                result.stdout_text, result.stderr_text
            )
        else:
            return result

    def spawn_expect(self, argv, default_timeout=10, encoding='utf-8',
                     extra_ssh_options=None):
        """Run command on remote host using IpaTestExpect"""
        return self.transport.spawn_expect(argv, default_timeout, encoding,
                                           extra_ssh_options)

class WinHost(pytest_multihost.host.WinHost):
    """
    Representation of a remote Windows host.

    This serves as a sketch class once we move from manual preparation of
    Active Directory to the automated setup.
    """
    transport_class = IPAOpenSSHTransport




import contextlib
import pytest

from cryptography import x509
import six

from ipapython.dn import DN, RDN, AVA, str2dn, dn2str, DECODING_ERROR
from ipapython import dn_ctypes


if six.PY3:
    unicode = str

    def cmp(a, b):
        if a == b:
            assert not a < b
            assert not a > b
            assert not a != b
            assert a <= b
            assert a >= b
            return 0
        elif a < b:
            assert not a > b
            assert a != b
            assert a <= b
            assert not a >= b
            return -1
        else:
            assert a > b
            assert a != b
            assert not a <= b
            assert a >= b
            return 1

pytestmark = pytest.mark.tier0


def expected_class(klass, component):
    if klass is AVA:
        if component == 'self':
            return AVA

    elif klass is RDN:
        if component == 'self':
            return RDN
        elif component == 'AVA':
            return AVA

    elif klass is DN:
        if component == 'self':
            return DN
        elif component == 'AVA':
            return AVA
        elif component == 'RDN':
            return RDN

    raise ValueError("class %s with component '%s' unknown" % (klass.__name__, component))


class TestAVA:
    @pytest.fixture(autouse=True)
    def ava_setup(self):
        self.attr1    = 'cn'
        self.value1   = 'Bob'
        self.str_ava1 = '%s=%s' % (self.attr1, self.value1)
        self.ava1     = AVA(self.attr1, self.value1)

        self.attr2    = 'ou'
        self.value2   = 'People'
        self.str_ava2 = '%s=%s' % (self.attr2, self.value2)
        self.ava2     = AVA(self.attr2, self.value2)

        self.attr3    = 'c'
        self.value3   = 'US'
        self.str_ava3 = '%s=%s' % (self.attr3, self.value3)
        self.ava3     = AVA(self.attr3, self.value3)

    def assertExpectedClass(self, klass, obj, component):
        assert obj.__class__ is expected_class(klass, component)

    def test_create(self):
        # Create with attr,value pair
        ava1 = AVA(self.attr1, self.value1)
        self.assertExpectedClass(AVA, ava1, 'self')
        assert ava1 == self.ava1

        # Create with "attr=value" string
        ava1 = AVA(self.str_ava1)
        self.assertExpectedClass(AVA, ava1, 'self')
        assert ava1 == self.ava1

        # Create with tuple (attr, value)
        ava1 = AVA((self.attr1, self.value1))
        self.assertExpectedClass(AVA, ava1, 'self')
        assert ava1 == self.ava1

        # Create with list [attr, value]
        ava1 = AVA([self.attr1, self.value1])
        self.assertExpectedClass(AVA, ava1, 'self')
        assert ava1 == self.ava1

        # Create with no args should fail
        with pytest.raises(TypeError):
            AVA()

        # Create with more than 3 args should fail
        with pytest.raises(TypeError):
            AVA(self.attr1, self.value1, self.attr1, self.attr1)

        # Create with 1 arg which is not string should fail
        with pytest.raises(TypeError):
            AVA(1)

        # Create with malformed AVA string should fail
        with pytest.raises(ValueError):
            AVA("cn")

        # Create with non-string parameters, should convert
        ava1 = AVA(1, self.value1)
        self.assertExpectedClass(AVA, ava1, 'self')
        assert ava1.attr == u'1'

        ava1 = AVA((1, self.value1))
        self.assertExpectedClass(AVA, ava1, 'self')
        assert ava1.attr == u'1'

        ava1 = AVA(self.attr1, 1)
        self.assertExpectedClass(AVA, ava1, 'self')
        assert ava1.value == u'1'

        ava1 = AVA((self.attr1, 1))
        self.assertExpectedClass(AVA, ava1, 'self')
        assert ava1.value == u'1'

    def test_indexing(self):
        ava1 = AVA(self.ava1)

        assert ava1[self.attr1] == self.value1

        assert ava1[0] == self.attr1
        assert ava1[1] == self.value1

        with pytest.raises(KeyError):
            ava1['foo']  # pylint: disable=pointless-statement

        with pytest.raises(KeyError):
            ava1[3]  # pylint: disable=pointless-statement

    def test_properties(self):
        ava1 = AVA(self.ava1)

        assert ava1.attr == self.attr1
        assert isinstance(ava1.attr, unicode)

        assert ava1.value == self.value1
        assert isinstance(ava1.value, unicode)

    def test_str(self):
        ava1 = AVA(self.ava1)

        assert str(ava1) == self.str_ava1
        assert isinstance(str(ava1), str)

    def test_cmp(self):
        # Equality
        ava1 = AVA(self.attr1, self.value1)

        assert ava1 == self.ava1
        assert ava1 == self.ava1

        assert ava1 == self.str_ava1
        assert ava1 == self.str_ava1

        result = cmp(ava1, self.ava1)
        assert result == 0

        # Upper case attr should still be equal
        ava1 = AVA(self.attr1.upper(), self.value1)

        assert ava1.attr != self.attr1
        assert ava1.value == self.value1
        assert ava1 == self.ava1
        assert ava1 == self.ava1

        result = cmp(ava1, self.ava1)
        assert result == 0

        # Upper case value should still be equal
        ava1 = AVA(self.attr1, self.value1.upper())

        assert ava1.attr == self.attr1
        assert ava1.value != self.value1
        assert ava1 == self.ava1
        assert ava1 == self.ava1

        result = cmp(ava1, self.ava1)
        assert result == 0

        # Make ava1's attr greater
        with pytest.raises(AttributeError):
            ava1.attr = self.attr1 + "1"
        ava1 = AVA(self.attr1 + "1", self.value1.upper())

        assert ava1 != self.ava1
        assert ava1 != self.ava1

        result = cmp(ava1, self.ava1)
        assert result == 1

        result = cmp(self.ava1, ava1)
        assert result == -1

        # Reset ava1's attr, should be equal again
        with pytest.raises(AttributeError):
            ava1.attr = self.attr1
        ava1 = AVA(self.attr1, self.value1.upper())

        result = cmp(ava1, self.ava1)
        assert result == 0

        # Make ava1's value greater
        # attr will be equal, this tests secondary comparision component
        with pytest.raises(AttributeError):
            ava1.value = self.value1 + "1"
        ava1 = AVA(self.attr1, self.value1 + "1")

        result = cmp(ava1, self.ava1)
        assert result == 1

        result = cmp(self.ava1, ava1)
        assert result == -1

    def test_hashing(self):
        # create AVA's that are equal but differ in case
        ava1 = AVA((self.attr1.lower(), self.value1.upper()))
        ava2 = AVA((self.attr1.upper(), self.value1.lower()))

        # AVAs that are equal should hash to the same value.
        assert ava1 == ava2
        assert hash(ava1) == hash(ava2)

        # Different AVA objects with the same value should
        # map to 1 common key and 1 member in a set. The key and
        # member are based on the object's value.

        ava1_a = AVA(self.ava1)
        ava1_b = AVA(self.ava1)

        ava2_a = AVA(self.ava2)
        ava2_b = AVA(self.ava2)

        ava3_a = AVA(self.ava3)
        ava3_b = AVA(self.ava3)

        assert ava1_a == ava1_b
        assert ava2_a == ava2_b
        assert ava3_a == ava3_b

        d = dict()
        s = set()

        d[ava1_a] = str(ava1_a)
        d[ava1_b] = str(ava1_b)
        d[ava2_a] = str(ava2_a)
        d[ava2_b] = str(ava2_b)

        s.add(ava1_a)
        s.add(ava1_b)
        s.add(ava2_a)
        s.add(ava2_b)

        assert len(d) == 2
        assert len(s) == 2
        assert sorted(d) == sorted([ava1_a, ava2_a])
        assert sorted(s) == sorted([ava1_a, ava2_a])

        assert ava1_a in d
        assert ava1_b in d
        assert ava2_a in d
        assert ava2_b in d
        assert ava3_a not in d
        assert ava3_b not in d

        assert ava1_a in s
        assert ava1_b in s
        assert ava2_a in s
        assert ava2_b in s
        assert ava3_a not in s
        assert ava3_b not in s


class TestRDN:
    @pytest.fixture(autouse=True)
    def rdn_setup(self):
        # ava1 must sort before ava2
        self.attr1    = 'cn'
        self.value1   = 'Bob'
        self.str_ava1 = '%s=%s' % (self.attr1, self.value1)
        self.ava1     = AVA(self.attr1, self.value1)

        self.str_rdn1 = '%s=%s' % (self.attr1, self.value1)
        self.rdn1 = RDN((self.attr1, self.value1))

        self.attr2    = 'ou'
        self.value2   = 'people'
        self.str_ava2 = '%s=%s' % (self.attr2, self.value2)
        self.ava2     = AVA(self.attr2, self.value2)

        self.str_rdn2 = '%s=%s' % (self.attr2, self.value2)
        self.rdn2     = RDN((self.attr2, self.value2))

        self.str_ava3 = '%s=%s+%s=%s' % (self.attr1, self.value1, self.attr2, self.value2)

        self.str_rdn3 = '%s=%s+%s=%s' % (self.attr1, self.value1, self.attr2, self.value2)
        self.rdn3 = RDN(self.ava1, self.ava2)

    def assertExpectedClass(self, klass, obj, component):
        assert obj.__class__ is expected_class(klass, component)

    def test_create(self):
        # Create with single attr,value pair
        rdn1 = RDN((self.attr1, self.value1))

        assert len(rdn1) == 1
        assert rdn1 == self.rdn1
        self.assertExpectedClass(RDN, rdn1, 'self')
        for i in range(0, len(rdn1)):
            self.assertExpectedClass(RDN, rdn1[i], 'AVA')
        assert rdn1[0] == self.ava1

        # Create with multiple attr,value pairs
        rdn3 = RDN((self.attr1, self.value1), (self.attr2, self.value2))
        assert len(rdn3) == 2
        assert rdn3 == self.rdn3
        self.assertExpectedClass(RDN, rdn3, 'self')
        for i in range(0, len(rdn3)):
            self.assertExpectedClass(RDN, rdn3[i], 'AVA')
        assert rdn3[0] == self.ava1
        assert rdn3[1] == self.ava2

        # Create with multiple attr,value pairs passed as lists
        rdn3 = RDN([self.attr1, self.value1], [self.attr2, self.value2])
        assert len(rdn3) == 2
        assert rdn3 == self.rdn3
        self.assertExpectedClass(RDN, rdn3, 'self')
        for i in range(0, len(rdn3)):
            self.assertExpectedClass(RDN, rdn3[i], 'AVA')
        assert rdn3[0] == self.ava1
        assert rdn3[1] == self.ava2

        # Create with multiple attr,value pairs but reverse
        # constructor parameter ordering. RDN canonical ordering
        # should remain the same
        rdn3 = RDN((self.attr2, self.value2), (self.attr1, self.value1))
        assert len(rdn3) == 2
        assert rdn3 == self.rdn3
        self.assertExpectedClass(RDN, rdn3, 'self')
        for i in range(0, len(rdn3)):
            self.assertExpectedClass(RDN, rdn3[i], 'AVA')
        assert rdn3[0] == self.ava1
        assert rdn3[1] == self.ava2

        # Create with single AVA object
        rdn1 = RDN(self.ava1)
        assert len(rdn1) == 1
        assert rdn1 == self.rdn1
        self.assertExpectedClass(RDN, rdn1, 'self')
        for i in range(0, len(rdn1)):
            self.assertExpectedClass(RDN, rdn1[i], 'AVA')
        assert rdn1[0] == self.ava1

        # Create with multiple AVA objects
        rdn3 = RDN(self.ava1, self.ava2)
        assert len(rdn3) == 2
        assert rdn3 == self.rdn3
        self.assertExpectedClass(RDN, rdn3, 'self')
        for i in range(0, len(rdn3)):
            self.assertExpectedClass(RDN, rdn3[i], 'AVA')
        assert rdn3[0] == self.ava1
        assert rdn3[1] == self.ava2


        # Create with multiple AVA objects but reverse constructor
        # parameter ordering.  RDN canonical ordering should remain
        # the same
        rdn3 = RDN(self.ava2, self.ava1)
        assert len(rdn3) == 2
        assert rdn3 == self.rdn3
        self.assertExpectedClass(RDN, rdn3, 'self')
        for i in range(0, len(rdn3)):
            self.assertExpectedClass(RDN, rdn3[i], 'AVA')
        assert rdn3[0] == self.ava1
        assert rdn3[1] == self.ava2

        # Create with single string with 1 AVA
        rdn1 = RDN(self.str_rdn1)
        assert len(rdn1) == 1
        assert rdn1 == self.rdn1
        self.assertExpectedClass(RDN, rdn1, 'self')
        for i in range(0, len(rdn1)):
            self.assertExpectedClass(RDN, rdn1[i], 'AVA')
        assert rdn1[0] == self.ava1

        # Create with single string with 2 AVA's
        rdn3 = RDN(self.str_rdn3)
        assert len(rdn3) == 2
        assert rdn3 == self.rdn3
        self.assertExpectedClass(RDN, rdn3, 'self')
        for i in range(0, len(rdn3)):
            self.assertExpectedClass(RDN, rdn3[i], 'AVA')
        assert rdn3[0] == self.ava1
        assert rdn3[1] == self.ava2

    def test_properties(self):
        rdn1 = RDN(self.rdn1)
        rdn2 = RDN(self.rdn2)
        rdn3 = RDN(self.rdn3)

        assert rdn1.attr == self.attr1
        assert isinstance(rdn1.attr, unicode)

        assert rdn1.value == self.value1
        assert isinstance(rdn1.value, unicode)

        assert rdn2.attr == self.attr2
        assert isinstance(rdn2.attr, unicode)

        assert rdn2.value == self.value2
        assert isinstance(rdn2.value, unicode)

        assert rdn3.attr == self.attr1
        assert isinstance(rdn3.attr, unicode)

        assert rdn3.value == self.value1
        assert isinstance(rdn3.value, unicode)

    def test_str(self):
        rdn1 = RDN(self.rdn1)
        rdn2 = RDN(self.rdn2)
        rdn3 = RDN(self.rdn3)

        assert str(rdn1) == self.str_rdn1
        assert isinstance(str(rdn1), str)

        assert str(rdn2) == self.str_rdn2
        assert isinstance(str(rdn2), str)

        assert str(rdn3) == self.str_rdn3
        assert isinstance(str(rdn3), str)

    def test_cmp(self):
        # Equality
        rdn1 = RDN((self.attr1, self.value1))

        assert rdn1 == self.rdn1
        assert rdn1 == self.rdn1

        assert rdn1 == self.str_rdn1
        assert rdn1 == self.str_rdn1

        result = cmp(rdn1, self.rdn1)
        assert result == 0

        # Make rdn1's attr greater
        rdn1 = RDN((self.attr1 + "1", self.value1))

        assert rdn1 != self.rdn1
        assert rdn1 != self.rdn1

        result = cmp(rdn1, self.rdn1)
        assert result == 1

        result = cmp(self.rdn1, rdn1)
        assert result == -1

        # Reset rdn1's attr, should be equal again
        rdn1 = RDN((self.attr1, self.value1))

        result = cmp(rdn1, self.rdn1)
        assert result == 0

        # Make rdn1's value greater
        # attr will be equal, this tests secondary comparision component
        rdn1 = RDN((self.attr1, self.value1 + "1"))

        result = cmp(rdn1, self.rdn1)
        assert result == 1

        result = cmp(self.rdn1, rdn1)
        assert result == -1

        # Make sure rdn's with more ava's are greater
        result = cmp(self.rdn1, self.rdn3)
        assert result == -1
        result = cmp(self.rdn3, self.rdn1)
        assert result == 1

    def test_indexing(self):
        rdn1 = RDN(self.rdn1)
        rdn2 = RDN(self.rdn2)
        rdn3 = RDN(self.rdn3)

        assert rdn1[0] == self.ava1
        assert rdn1[self.ava1.attr] == self.ava1.value
        with pytest.raises(KeyError):
            rdn1['foo']  # pylint: disable=pointless-statement

        assert rdn2[0] == self.ava2
        assert rdn2[self.ava2.attr] == self.ava2.value
        with pytest.raises(KeyError):
            rdn2['foo']  # pylint: disable=pointless-statement

        assert rdn3[0] == self.ava1
        assert rdn3[self.ava1.attr] == self.ava1.value
        assert rdn3[1] == self.ava2
        assert rdn3[self.ava2.attr] == self.ava2.value
        with pytest.raises(KeyError):
            rdn3['foo']  # pylint: disable=pointless-statement

        assert rdn1.attr == self.attr1
        assert rdn1.value == self.value1

        with pytest.raises(TypeError):
            rdn3[1.0]  # pylint: disable=pointless-statement

        # Slices
        assert rdn3[0:1] == [self.ava1]
        assert rdn3[:] == [self.ava1, self.ava2]

    def test_assignments(self):
        rdn = RDN((self.attr1, self.value1))
        with pytest.raises(TypeError):
            # pylint: disable=unsupported-assignment-operation
            rdn[0] = self.ava2

    def test_iter(self):
        rdn1 = RDN(self.rdn1)
        rdn2 = RDN(self.rdn2)
        rdn3 = RDN(self.rdn3)

        assert len(rdn1) == 1
        assert rdn1[:] == [self.ava1]
        for i, ava in enumerate(rdn1):
            if i == 0:
                assert ava == self.ava1
            else:
                pytest.fail(
                    "got iteration index %d, but len=%d" % (i, len(rdn1)))

        assert len(rdn2) == 1
        assert rdn2[:] == [self.ava2]
        for i, ava in enumerate(rdn2):
            if i == 0:
                assert ava == self.ava2
            else:
                pytest.fail(
                    "got iteration index %d, but len=%d" % (i, len(rdn2)))

        assert len(rdn3) == 2
        assert rdn3[:] == [self.ava1, self.ava2]
        for i, ava in enumerate(rdn3):
            if i == 0:
                assert ava == self.ava1
            elif i == 1:
                assert ava == self.ava2
            else:
                pytest.fail(
                    "got iteration index %d, but len=%d" % (i, len(rdn3)))


    def test_concat(self):
        rdn1 = RDN((self.attr1, self.value1))
        rdn2 = RDN((self.attr2, self.value2))

        # in-place addtion
        rdn1 += rdn2
        assert rdn1 == self.rdn3
        self.assertExpectedClass(RDN, rdn1, 'self')
        for i in range(0, len(rdn1)):
            self.assertExpectedClass(RDN, rdn1[i], 'AVA')

        rdn1 = RDN((self.attr1, self.value1))
        rdn1 += self.ava2
        assert rdn1 == self.rdn3
        self.assertExpectedClass(RDN, rdn1, 'self')
        for i in range(0, len(rdn1)):
            self.assertExpectedClass(RDN, rdn1[i], 'AVA')

        rdn1 = RDN((self.attr1, self.value1))
        rdn1 += self.str_ava2
        assert rdn1 == self.rdn3
        self.assertExpectedClass(RDN, rdn1, 'self')
        for i in range(0, len(rdn1)):
            self.assertExpectedClass(RDN, rdn1[i], 'AVA')

        # concatenation
        rdn1 = RDN((self.attr1, self.value1))
        rdn3 = rdn1 + rdn2
        assert rdn3 == self.rdn3
        self.assertExpectedClass(RDN, rdn3, 'self')
        for i in range(0, len(rdn3)):
            self.assertExpectedClass(RDN, rdn3[i], 'AVA')

        rdn3 = rdn1 + self.ava2
        assert rdn3 == self.rdn3
        self.assertExpectedClass(RDN, rdn3, 'self')
        for i in range(0, len(rdn3)):
            self.assertExpectedClass(RDN, rdn3[i], 'AVA')

        rdn3 = rdn1 + self.str_ava2
        assert rdn3 == self.rdn3
        self.assertExpectedClass(RDN, rdn3, 'self')
        for i in range(0, len(rdn3)):
            self.assertExpectedClass(RDN, rdn3[i], 'AVA')


    def test_hashing(self):
        # create RDN's that are equal but differ in case
        rdn1 = RDN((self.attr1.lower(), self.value1.upper()))
        rdn2 = RDN((self.attr1.upper(), self.value1.lower()))

        # RDNs that are equal should hash to the same value.
        assert rdn1 == rdn2
        assert hash(rdn1) == hash(rdn2)


class TestDN:
    @pytest.fixture(autouse=True)
    def dn_setup(self):
        # ava1 must sort before ava2
        self.attr1    = 'cn'
        self.value1   = u'Bob'
        self.str_ava1 = '%s=%s' % (self.attr1, self.value1)
        self.ava1     = AVA(self.attr1, self.value1)

        self.str_rdn1 = '%s=%s' % (self.attr1, self.value1)
        self.rdn1     = RDN((self.attr1, self.value1))

        self.attr2    = 'ou'
        self.value2   = u'people'
        self.str_ava2 = '%s=%s' % (self.attr2, self.value2)
        self.ava2     = AVA(self.attr2, self.value2)

        self.str_rdn2 = '%s=%s' % (self.attr2, self.value2)
        self.rdn2     = RDN((self.attr2, self.value2))

        self.str_dn1 = self.str_rdn1
        self.dn1 = DN(self.rdn1)

        self.str_dn2 = self.str_rdn2
        self.dn2 = DN(self.rdn2)

        self.str_dn3 = '%s,%s' % (self.str_rdn1, self.str_rdn2)
        self.dn3 = DN(self.rdn1, self.rdn2)

        self.base_rdn1 = RDN(('dc', 'redhat'))
        self.base_rdn2 = RDN(('dc', 'com'))
        self.base_dn = DN(self.base_rdn1, self.base_rdn2)

        self.container_rdn1 = RDN(('cn', 'sudorules'))
        self.container_rdn2 = RDN(('cn', 'sudo'))
        self.container_dn = DN(self.container_rdn1, self.container_rdn2)

        self.base_container_dn = DN((self.attr1, self.value1),
                                    self.container_dn, self.base_dn)

        ou = x509.NameAttribute(
            x509.NameOID.ORGANIZATIONAL_UNIT_NAME, self.value2)
        cn = x509.NameAttribute(x509.NameOID.COMMON_NAME, self.value1)
        c = x509.NameAttribute(x509.NameOID.COUNTRY_NAME, 'AU')
        st = x509.NameAttribute(
            x509.NameOID.STATE_OR_PROVINCE_NAME, 'Queensland')
        self.x500name = x509.Name([ou, cn])
        self.x500nameMultiRDN = x509.Name([
            x509.RelativeDistinguishedName([c, st]),
            x509.RelativeDistinguishedName([cn]),
        ])
        self.x500nameMultiRDN2 = x509.Name([
            x509.RelativeDistinguishedName([st, c]),
            x509.RelativeDistinguishedName([cn]),
        ])

    def assertExpectedClass(self, klass, obj, component):
        assert obj.__class__ is expected_class(klass, component)

    def test_create(self):
        # Create with single attr,value pair
        dn1 = DN((self.attr1, self.value1))
        assert len(dn1) == 1
        self.assertExpectedClass(DN, dn1, 'self')
        for i in range(0, len(dn1)):
            self.assertExpectedClass(DN, dn1[i], 'RDN')
            for j in range(0, len(dn1[i])):
                self.assertExpectedClass(DN, dn1[i][j], 'AVA')
        assert isinstance(dn1[0].attr, unicode)
        assert isinstance(dn1[0].value, unicode)
        assert dn1[0] == self.rdn1

        # Create with single attr,value pair passed as a tuple
        dn1 = DN((self.attr1, self.value1))
        assert len(dn1) == 1
        self.assertExpectedClass(DN, dn1, 'self')
        for i in range(0, len(dn1)):
            self.assertExpectedClass(DN, dn1[i], 'RDN')
            for j in range(0, len(dn1[i])):
                self.assertExpectedClass(DN, dn1[i][j], 'AVA')
            assert isinstance(dn1[i].attr, unicode)
            assert isinstance(dn1[i].value, unicode)
        assert dn1[0] == self.rdn1

        # Creation with multiple attr,value string pairs should fail
        with pytest.raises(ValueError):
            dn1 = DN(self.attr1, self.value1, self.attr2, self.value2)

        # Create with multiple attr,value pairs passed as tuples & lists
        dn1 = DN((self.attr1, self.value1), [self.attr2, self.value2])
        assert len(dn1) == 2
        self.assertExpectedClass(DN, dn1, 'self')
        for i in range(0, len(dn1)):
            self.assertExpectedClass(DN, dn1[i], 'RDN')
            for j in range(0, len(dn1[i])):
                self.assertExpectedClass(DN, dn1[i][j], 'AVA')
            assert isinstance(dn1[i].attr, unicode)
            assert isinstance(dn1[i].value, unicode)
        assert dn1[0] == self.rdn1
        assert dn1[1] == self.rdn2

        # Create with multiple attr,value pairs passed as tuple and RDN
        dn1 = DN((self.attr1, self.value1), RDN((self.attr2, self.value2)))
        assert len(dn1) == 2
        self.assertExpectedClass(DN, dn1, 'self')
        for i in range(0, len(dn1)):
            self.assertExpectedClass(DN, dn1[i], 'RDN')
            for j in range(0, len(dn1[i])):
                self.assertExpectedClass(DN, dn1[i][j], 'AVA')
            assert isinstance(dn1[i].attr, unicode)
            assert isinstance(dn1[i].value, unicode)
        assert dn1[0] == self.rdn1
        assert dn1[1] == self.rdn2

        # Create with multiple attr,value pairs but reverse
        # constructor parameter ordering. RDN ordering should also be
        # reversed because DN's are a ordered sequence of RDN's
        dn1 = DN((self.attr2, self.value2), (self.attr1, self.value1))
        assert len(dn1) == 2
        self.assertExpectedClass(DN, dn1, 'self')
        for i in range(0, len(dn1)):
            self.assertExpectedClass(DN, dn1[i], 'RDN')
            for j in range(0, len(dn1[i])):
                self.assertExpectedClass(DN, dn1[i][j], 'AVA')
            assert isinstance(dn1[i].attr, unicode)
            assert isinstance(dn1[i].value, unicode)
        assert dn1[0] == self.rdn2
        assert dn1[1] == self.rdn1

        # Create with single RDN object
        dn1 = DN(self.rdn1)
        assert len(dn1) == 1
        self.assertExpectedClass(DN, dn1, 'self')
        for i in range(0, len(dn1)):
            self.assertExpectedClass(DN, dn1[i], 'RDN')
            for j in range(0, len(dn1[i])):
                self.assertExpectedClass(DN, dn1[i][j], 'AVA')
            assert isinstance(dn1[i].attr, unicode)
            assert isinstance(dn1[i].value, unicode)
        assert dn1[0] == self.rdn1

        # Create with multiple RDN objects, assure ordering is preserved.
        dn1 = DN(self.rdn1, self.rdn2)
        assert len(dn1) == 2
        self.assertExpectedClass(DN, dn1, 'self')
        for i in range(0, len(dn1)):
            self.assertExpectedClass(DN, dn1[i], 'RDN')
            for j in range(0, len(dn1[i])):
                self.assertExpectedClass(DN, dn1[i][j], 'AVA')
            assert isinstance(dn1[i].attr, unicode)
            assert isinstance(dn1[i].value, unicode)
        assert dn1[0] == self.rdn1
        assert dn1[1] == self.rdn2

        # Create with multiple RDN objects in different order, assure
        # ordering is preserved.
        dn1 = DN(self.rdn2, self.rdn1)
        assert len(dn1) == 2
        self.assertExpectedClass(DN, dn1, 'self')
        for i in range(0, len(dn1)):
            self.assertExpectedClass(DN, dn1[i], 'RDN')
            for j in range(0, len(dn1[i])):
                self.assertExpectedClass(DN, dn1[i][j], 'AVA')
            assert isinstance(dn1[i].attr, unicode)
            assert isinstance(dn1[i].value, unicode)
        assert dn1[0] == self.rdn2
        assert dn1[1] == self.rdn1

        # Create with single string with 1 RDN
        dn1 = DN(self.str_rdn1)
        assert len(dn1) == 1
        self.assertExpectedClass(DN, dn1, 'self')
        for i in range(0, len(dn1)):
            self.assertExpectedClass(DN, dn1[i], 'RDN')
            for j in range(0, len(dn1[i])):
                self.assertExpectedClass(DN, dn1[i][j], 'AVA')
            assert isinstance(dn1[i].attr, unicode)
            assert isinstance(dn1[i].value, unicode)
        assert dn1[0] == self.rdn1

        # Create with single string with 2 RDN's
        dn1 = DN(self.str_dn3)
        assert len(dn1) == 2
        self.assertExpectedClass(DN, dn1, 'self')
        for i in range(0, len(dn1)):
            self.assertExpectedClass(DN, dn1[i], 'RDN')
            for j in range(0, len(dn1[i])):
                self.assertExpectedClass(DN, dn1[i][j], 'AVA')
            assert isinstance(dn1[i].attr, unicode)
            assert isinstance(dn1[i].value, unicode)
        assert dn1[0] == self.rdn1
        assert dn1[1] == self.rdn2

        # Create with a python-cryptography 'Name'
        dn1 = DN(self.x500name)
        assert len(dn1) == 2
        self.assertExpectedClass(DN, dn1, 'self')
        for i in range(0, len(dn1)):
            self.assertExpectedClass(DN, dn1[i], 'RDN')
            for j in range(0, len(dn1[i])):
                self.assertExpectedClass(DN, dn1[i][j], 'AVA')
            assert isinstance(dn1[i].attr, unicode)
            assert isinstance(dn1[i].value, unicode)
        assert dn1[0] == self.rdn1
        assert dn1[1] == self.rdn2

        # Create from 'Name' with multi-valued RDN
        dn1 = DN(self.x500nameMultiRDN)
        assert len(dn1) == 2
        assert len(dn1[1]) == 2
        assert AVA('c', 'au') in dn1[1]
        assert AVA('st', 'queensland') in dn1[1]
        assert len(dn1[0]) == 1
        assert self.ava1 in dn1[0]

        # Create with RDN, and 2 DN's (e.g. attr + container + base)
        dn1 = DN((self.attr1, self.value1), self.container_dn, self.base_dn)
        assert len(dn1) == 5
        dn_str = ','.join([str(self.rdn1),
                            str(self.container_rdn1), str(self.container_rdn2),
                            str(self.base_rdn1), str(self.base_rdn2)])
        assert str(dn1) == dn_str

    def test_str(self):
        dn1 = DN(self.dn1)
        dn2 = DN(self.dn2)
        dn3 = DN(self.dn3)

        assert str(dn1) == self.str_dn1
        assert isinstance(str(dn1), str)

        assert str(dn2) == self.str_dn2
        assert isinstance(str(dn2), str)

        assert str(dn3) == self.str_dn3
        assert isinstance(str(dn3), str)

    def test_cmp(self):
        # Equality
        dn1 = DN((self.attr1, self.value1))

        assert dn1 == self.dn1
        assert dn1 == self.dn1

        assert dn1 == self.str_dn1
        assert dn1 == self.str_dn1

        result = cmp(dn1, self.dn1)
        assert result == 0

        # Make dn1's attr greater
        with pytest.raises(AttributeError):
            dn1[0].attr = self.attr1 + "1"
        dn1 = DN((self.attr1 + "1", self.value1))

        assert dn1 != self.dn1
        assert dn1 != self.dn1

        result = cmp(dn1, self.dn1)
        assert result == 1

        result = cmp(self.dn1, dn1)
        assert result == -1

        # Reset dn1's attr, should be equal again
        with pytest.raises(AttributeError):
            dn1[0].attr = self.attr1
        dn1 = DN((self.attr1, self.value1))

        result = cmp(dn1, self.dn1)
        assert result == 0

        # Make dn1's value greater
        # attr will be equal, this tests secondary comparision component
        with pytest.raises(AttributeError):
            dn1[0].value = self.value1 + "1"
        dn1 = DN((self.attr1, self.value1 + "1"))

        result = cmp(dn1, self.dn1)
        assert result == 1

        result = cmp(self.dn1, dn1)
        assert result == -1

        # Make sure dn's with more rdn's are greater
        result = cmp(self.dn1, self.dn3)
        assert result == -1
        result = cmp(self.dn3, self.dn1)
        assert result == 1


        # Test startswith, endswith
        container_dn = DN(self.container_dn)
        base_container_dn = DN(self.base_container_dn)

        assert base_container_dn.startswith(self.rdn1)
        assert base_container_dn.startswith(self.dn1)
        assert base_container_dn.startswith(self.dn1 + container_dn)
        assert not base_container_dn.startswith(self.dn2)
        assert not base_container_dn.startswith(self.rdn2)
        assert base_container_dn.startswith((self.dn1))
        assert base_container_dn.startswith((self.rdn1))
        assert not base_container_dn.startswith((self.rdn2))
        assert base_container_dn.startswith((self.rdn2, self.rdn1))
        assert base_container_dn.startswith((self.dn1, self.dn2))

        assert base_container_dn.endswith(self.base_dn)
        assert base_container_dn.endswith(container_dn + self.base_dn)
        assert not base_container_dn.endswith(DN(self.base_rdn1))
        assert base_container_dn.endswith(DN(self.base_rdn2))
        assert base_container_dn.endswith(
            (DN(self.base_rdn1), DN(self.base_rdn2)))

        # Test "in" membership
        assert self.container_rdn1 in container_dn
        # pylint: disable=comparison-with-itself
        assert container_dn in container_dn
        assert self.base_rdn1 not in container_dn

        assert self.container_rdn1 in base_container_dn
        assert container_dn in base_container_dn
        assert container_dn + self.base_dn in base_container_dn
        assert self.dn1 + container_dn + self.base_dn in base_container_dn
        assert self.dn1 + container_dn + self.base_dn == base_container_dn

        assert self.container_rdn1 not in self.base_dn

    def test_eq_multi_rdn(self):
        dn1 = DN(self.ava1, 'ST=Queensland+C=AU')
        dn2 = DN(self.ava1, 'C=AU+ST=Queensland')
        assert dn1 == dn2

        # ensure AVAs get sorted when constructing from x509.Name
        dn3 = DN(self.x500nameMultiRDN)
        dn4 = DN(self.x500nameMultiRDN2)
        assert dn3 == dn4

        # ensure AVAs get sorted in the same way regardless of what
        # the DN was constructed from
        assert dn1 == dn3
        assert dn1 == dn4
        assert dn2 == dn3
        assert dn2 == dn4

    def test_indexing(self):
        dn1 = DN(self.dn1)
        dn2 = DN(self.dn2)
        dn3 = DN(self.dn3)

        assert dn1[0] == self.rdn1
        assert dn1[self.rdn1.attr] == self.rdn1.value
        with pytest.raises(KeyError):
            dn1['foo']  # pylint: disable=pointless-statement

        assert dn2[0] == self.rdn2
        assert dn2[self.rdn2.attr] == self.rdn2.value
        with pytest.raises(KeyError):
            dn2['foo']  # pylint: disable=pointless-statement

        assert dn3[0] == self.rdn1
        assert dn3[self.rdn1.attr] == self.rdn1.value
        assert dn3[1] == self.rdn2
        assert dn3[self.rdn2.attr] == self.rdn2.value
        with pytest.raises(KeyError):
            dn3['foo']  # pylint: disable=pointless-statement

        with pytest.raises(TypeError):
            dn3[1.0]  # pylint: disable=pointless-statement

    def test_assignments(self):
        dn = DN('t=0,t=1,t=2,t=3,t=4,t=5,t=6,t=7,t=8,t=9')
        with pytest.raises(TypeError):
            # pylint: disable=unsupported-assignment-operation
            dn[0] = RDN('t=a')
        with pytest.raises(TypeError):
            # pylint: disable=unsupported-assignment-operation
            dn[0:1] = [RDN('t=a'), RDN('t=b')]

    def test_iter(self):
        dn1 = DN(self.dn1)
        dn2 = DN(self.dn2)
        dn3 = DN(self.dn3)

        assert len(dn1) == 1
        assert dn1[:] == self.rdn1
        for i, ava in enumerate(dn1):
            if i == 0:
                assert ava == self.rdn1
            else:
                pytest.fail(
                    "got iteration index %d, but len=%d" % (i, len(self.rdn1)))

        assert len(dn2) == 1
        assert dn2[:] == self.rdn2
        for i, ava in enumerate(dn2):
            if i == 0:
                assert ava == self.rdn2
            else:
                pytest.fail(
                    "got iteration index %d, but len=%d" % (i, len(self.rdn2)))

        assert len(dn3) == 2
        assert dn3[:] == DN(self.rdn1, self.rdn2)
        for i, ava in enumerate(dn3):
            if i == 0:
                assert ava == self.rdn1
            elif i == 1:
                assert ava == self.rdn2
            else:
                pytest.fail(
                    "got iteration index %d, but len=%d" % (i, len(dn3)))

    def test_concat(self):
        dn1 = DN((self.attr1, self.value1))
        dn2 = DN([self.attr2, self.value2])

        # in-place addtion

        dn1 += dn2
        assert dn1 == self.dn3
        self.assertExpectedClass(DN, dn1, 'self')
        for i in range(0, len(dn1)):
            self.assertExpectedClass(DN, dn1[i], 'RDN')
            for j in range(0, len(dn1[i])):
                self.assertExpectedClass(DN, dn1[i][j], 'AVA')


        dn1 = DN((self.attr1, self.value1))
        dn1 += self.rdn2
        assert dn1 == self.dn3
        self.assertExpectedClass(DN, dn1, 'self')
        for i in range(0, len(dn1)):
            self.assertExpectedClass(DN, dn1[i], 'RDN')
            for j in range(0, len(dn1[i])):
                self.assertExpectedClass(DN, dn1[i][j], 'AVA')


        dn1 = DN((self.attr1, self.value1))
        dn1 += self.dn2
        assert dn1 == self.dn3
        self.assertExpectedClass(DN, dn1, 'self')
        for i in range(0, len(dn1)):
            self.assertExpectedClass(DN, dn1[i], 'RDN')
            for j in range(0, len(dn1[i])):
                self.assertExpectedClass(DN, dn1[i][j], 'AVA')


        dn1 = DN((self.attr1, self.value1))
        dn1 += self.str_dn2
        assert dn1 == self.dn3
        self.assertExpectedClass(DN, dn1, 'self')
        for i in range(0, len(dn1)):
            self.assertExpectedClass(DN, dn1[i], 'RDN')
            for j in range(0, len(dn1[i])):
                self.assertExpectedClass(DN, dn1[i][j], 'AVA')


        # concatenation
        dn1 = DN((self.attr1, self.value1))
        dn3 = dn1 + dn2
        assert dn3 == self.dn3
        self.assertExpectedClass(DN, dn3, 'self')
        for i in range(0, len(dn3)):
            self.assertExpectedClass(DN, dn3[i], 'RDN')
            for j in range(0, len(dn3[i])):
                self.assertExpectedClass(DN, dn3[i][j], 'AVA')


        dn1 = DN((self.attr1, self.value1))
        dn3 = dn1 + self.rdn2
        assert dn3 == self.dn3
        self.assertExpectedClass(DN, dn3, 'self')
        for i in range(0, len(dn3)):
            self.assertExpectedClass(DN, dn3[i], 'RDN')
            for j in range(0, len(dn3[i])):
                self.assertExpectedClass(DN, dn3[i][j], 'AVA')

        dn3 = dn1 + self.str_rdn2
        assert dn3 == self.dn3
        self.assertExpectedClass(DN, dn3, 'self')
        for i in range(0, len(dn3)):
            self.assertExpectedClass(DN, dn3[i], 'RDN')
            self.assertExpectedClass(DN, dn3[i][0], 'AVA')

        dn3 = dn1 + self.str_dn2
        assert dn3 == self.dn3
        self.assertExpectedClass(DN, dn3, 'self')
        self.assertExpectedClass(DN, dn3, 'self')
        for i in range(0, len(dn3)):
            self.assertExpectedClass(DN, dn3[i], 'RDN')
            for j in range(0, len(dn3[i])):
                self.assertExpectedClass(DN, dn3[i][j], 'AVA')

        dn3 = dn1 + self.dn2
        assert dn3 == self.dn3
        self.assertExpectedClass(DN, dn3, 'self')
        self.assertExpectedClass(DN, dn3, 'self')
        for i in range(0, len(dn3)):
            self.assertExpectedClass(DN, dn3[i], 'RDN')
            for j in range(0, len(dn3[i])):
                self.assertExpectedClass(DN, dn3[i][j], 'AVA')

    def test_find(self):
        #        -10 -9  -8     -7  -6  -5  -4     -3  -2  -1
        dn = DN('t=0,t=1,cn=bob,t=3,t=4,t=5,cn=bob,t=7,t=8,t=9')
        pat = DN('cn=bob')

        # forward
        assert dn.find(pat) == 2
        assert dn.find(pat, 1) == 2
        assert dn.find(pat, 1, 3) == 2
        assert dn.find(pat, 2, 3) == 2
        assert dn.find(pat, 6) == 6

        assert dn.find(pat, 7) == -1
        assert dn.find(pat, 1, 2) == -1

        with pytest.raises(ValueError):
            assert dn.index(pat, 7) == -1
        with pytest.raises(ValueError):
            assert dn.index(pat, 1, 2) == -1

        # reverse
        assert dn.rfind(pat) == 6
        assert dn.rfind(pat, -4) == 6
        assert dn.rfind(pat, 6) == 6
        assert dn.rfind(pat, 6, 8) == 6
        assert dn.rfind(pat, 6, 8) == 6
        assert dn.rfind(pat, -8) == 6
        assert dn.rfind(pat, -8, -4) == 6
        assert dn.rfind(pat, -8, -5) == 2

        assert dn.rfind(pat, 7) == -1
        assert dn.rfind(pat, -3) == -1

        with pytest.raises(ValueError):
            assert dn.rindex(pat, 7) == -1
        with pytest.raises(ValueError):
            assert dn.rindex(pat, -3) == -1

    def test_replace(self):
        # pylint: disable=no-member
        dn = DN('t=0,t=1,t=2,t=3,t=4,t=5,t=6,t=7,t=8,t=9')
        with pytest.raises(AttributeError):
            dn.replace  # pylint: disable=pointless-statement

    def test_hashing(self):
        # create DN's that are equal but differ in case
        dn1 = DN((self.attr1.lower(), self.value1.upper()))
        dn2 = DN((self.attr1.upper(), self.value1.lower()))

        # DNs that are equal should hash to the same value.
        assert dn1 == dn2

        # Good, everyone's equal, now verify their hash values

        assert hash(dn1) == hash(dn2)

        # Different DN objects with the same value should
        # map to 1 common key and 1 member in a set. The key and
        # member are based on the object's value.

        dn1_a = DN(self.dn1)
        dn1_b = DN(self.dn1)

        dn2_a = DN(self.dn2)
        dn2_b = DN(self.dn2)

        dn3_a = DN(self.dn3)
        dn3_b = DN(self.dn3)

        assert dn1_a == dn1_b
        assert dn2_a == dn2_b
        assert dn3_a == dn3_b

        d = dict()
        s = set()

        d[dn1_a] = str(dn1_a)
        d[dn1_b] = str(dn1_b)
        d[dn2_a] = str(dn2_a)
        d[dn2_b] = str(dn2_b)

        s.add(dn1_a)
        s.add(dn1_b)
        s.add(dn2_a)
        s.add(dn2_b)

        assert len(d) == 2
        assert len(s) == 2
        assert sorted(d) == sorted([dn1_a, dn2_a])
        assert sorted(s) == sorted([dn1_a, dn2_a])

        assert dn1_a in d
        assert dn1_b in d
        assert dn2_a in d
        assert dn2_b in d
        assert dn3_a not in d
        assert dn3_b not in d

        assert dn1_a in s
        assert dn1_b in s
        assert dn2_a in s
        assert dn2_b in s
        assert dn3_a not in s
        assert dn3_b not in s

    def test_x500_text(self):
        # null DN x500 ordering and LDAP ordering are the same
        nulldn = DN()
        assert nulldn.ldap_text() == nulldn.x500_text()

        # reverse a DN with a single RDN
        assert self.dn1.ldap_text() == self.dn1.x500_text()

        # reverse a DN with 2 RDNs
        dn3_x500 = self.dn3.x500_text()
        dn3_rev = DN(self.rdn2, self.rdn1)
        assert dn3_rev.ldap_text() == dn3_x500

        # reverse a longer DN
        longdn_x500 = self.base_container_dn.x500_text()
        longdn_rev = DN(longdn_x500)
        l = len(self.base_container_dn)
        for i in range(l):
            assert longdn_rev[i] == self.base_container_dn[l - 1 - i]


class TestEscapes:
    @pytest.fixture(autouse=True)
    def escapes_setup(self):
        self.privilege = 'R,W privilege'
        self.dn_str_hex_escape = 'cn=R\\2cW privilege,cn=privileges,cn=pbac,dc=idm,dc=lab,dc=bos,dc=redhat,dc=com'
        self.dn_str_backslash_escape = 'cn=R\\,W privilege,cn=privileges,cn=pbac,dc=idm,dc=lab,dc=bos,dc=redhat,dc=com'

    def test_escape(self):
        dn = DN(self.dn_str_hex_escape)
        assert dn['cn'] == self.privilege
        assert dn[0].value == self.privilege

        dn = DN(self.dn_str_backslash_escape)
        assert dn['cn'] == self.privilege
        assert dn[0].value == self.privilege


class TestInternationalization:
    @pytest.fixture(autouse=True)
    def i18n_setup(self):
        # Hello in Arabic
        self.arabic_hello_utf8 = (b'\xd9\x85\xd9\x83\xd9\x8a\xd9\x84' +
                                  b'\xd8\xb9\x20\xd9\x85\xd8\xa7\xd9' +
                                  b'\x84\xd9\x91\xd8\xb3\xd9\x84\xd8\xa7')

        self.arabic_hello_unicode = self.arabic_hello_utf8.decode('utf-8')

    def assert_equal_utf8(self, obj, b):
        if six.PY2:
            assert str(obj) == b
        else:
            assert str(obj) == b.decode('utf-8')

    @contextlib.contextmanager
    def fail_py3(self, exception_type):
        try:
            yield
        except exception_type:
            if six.PY2:
                raise

    def test_i18n(self):
        actual = self.arabic_hello_unicode.encode('utf-8')
        expected = self.arabic_hello_utf8
        assert actual == expected

        # AVA's
        # test attr i18n
        ava1 = AVA(self.arabic_hello_unicode, 'foo')
        assert isinstance(ava1.attr, unicode)
        assert isinstance(ava1.value, unicode)
        assert ava1.attr == self.arabic_hello_unicode
        self.assert_equal_utf8(ava1, self.arabic_hello_utf8 + b'=foo')

        with self.fail_py3(TypeError):
            ava1 = AVA(self.arabic_hello_utf8, 'foo')
        if six.PY2:
            assert isinstance(ava1.attr, unicode)
            assert isinstance(ava1.value, unicode)
            assert ava1.attr == self.arabic_hello_unicode
            self.assert_equal_utf8(ava1, self.arabic_hello_utf8 + b'=foo')

        # test value i18n
        ava1 = AVA('cn', self.arabic_hello_unicode)
        assert isinstance(ava1.attr, unicode)
        assert isinstance(ava1.value, unicode)
        assert ava1.value == self.arabic_hello_unicode
        self.assert_equal_utf8(ava1, b'cn=' + self.arabic_hello_utf8)

        with self.fail_py3(TypeError):
            ava1 = AVA('cn', self.arabic_hello_utf8)
        if six.PY2:
            assert isinstance(ava1.attr, unicode)
            assert isinstance(ava1.value, unicode)
            assert ava1.value == self.arabic_hello_unicode
            self.assert_equal_utf8(ava1, b'cn=' + self.arabic_hello_utf8)

        # RDN's
        # test attr i18n
        rdn1 = RDN((self.arabic_hello_unicode, 'foo'))
        assert isinstance(rdn1.attr, unicode)
        assert isinstance(rdn1.value, unicode)
        assert rdn1.attr == self.arabic_hello_unicode
        self.assert_equal_utf8(rdn1, self.arabic_hello_utf8 + b'=foo')

        with self.fail_py3(TypeError):
            rdn1 = RDN((self.arabic_hello_utf8, 'foo'))
        if six.PY2:
            assert isinstance(rdn1.attr, unicode)
            assert isinstance(rdn1.value, unicode)
            assert rdn1.attr == self.arabic_hello_unicode
            assert str(rdn1) == self.arabic_hello_utf8 + b'=foo'

        # test value i18n
        rdn1 = RDN(('cn', self.arabic_hello_unicode))
        assert isinstance(rdn1.attr, unicode)
        assert isinstance(rdn1.value, unicode)
        assert rdn1.value == self.arabic_hello_unicode
        self.assert_equal_utf8(rdn1, b'cn=' + self.arabic_hello_utf8)

        with self.fail_py3(TypeError):
            rdn1 = RDN(('cn', self.arabic_hello_utf8))
        if six.PY2:
            assert isinstance(rdn1.attr, unicode)
            assert isinstance(rdn1.value, unicode)
            assert rdn1.value == self.arabic_hello_unicode
            assert str(rdn1) == b'cn=' + self.arabic_hello_utf8

        # DN's
        # test attr i18n
        dn1 = DN((self.arabic_hello_unicode, 'foo'))
        assert isinstance(dn1[0].attr, unicode)
        assert isinstance(dn1[0].value, unicode)
        assert dn1[0].attr == self.arabic_hello_unicode
        self.assert_equal_utf8(dn1, self.arabic_hello_utf8 + b'=foo')

        with self.fail_py3(TypeError):
            dn1 = DN((self.arabic_hello_utf8, 'foo'))
        if six.PY2:
            assert isinstance(dn1[0].attr, unicode)
            assert isinstance(dn1[0].value, unicode)
            assert dn1[0].attr == self.arabic_hello_unicode
            assert str(dn1) == self.arabic_hello_utf8 + b'=foo'

        # test value i18n
        dn1 = DN(('cn', self.arabic_hello_unicode))
        assert isinstance(dn1[0].attr, unicode)
        assert isinstance(dn1[0].value, unicode)
        assert dn1[0].value == self.arabic_hello_unicode
        self.assert_equal_utf8(dn1, b'cn=' + self.arabic_hello_utf8)

        with self.fail_py3(TypeError):
            dn1 = DN(('cn', self.arabic_hello_utf8))
        if six.PY2:
            assert isinstance(dn1[0].attr, unicode)
            assert isinstance(dn1[0].value, unicode)
            assert dn1[0].value == self.arabic_hello_unicode
            assert str(dn1) == b'cn=' + self.arabic_hello_utf8


# 1: LDAP_AVA_STRING
# 4: LDAP_AVA_NONPRINTABLE
@pytest.mark.parametrize(
    'dnstring,expected',
    [
        ('', []),
        ('cn=bob', [[('cn', 'bob', 1)]]),
        ('cn=Bob', [[('cn', 'Bob', 1)]]),
        (u'cn=b\xf6b', [[('cn', u'b\xf6b', 4)]]),
        ('cn=bob,sn=builder', [[('cn', 'bob', 1)], [('sn', 'builder', 1)]]),
        (u'cn=b\xf6b,sn=builder', [
            [('cn', u'b\xf6b', 4)], [('sn', 'builder', 1)]
        ]),
        ('cn=bob+sn=builder', [[('cn', 'bob', 1), ('sn', 'builder', 1)]]),
        ('dc=ipa,dc=example', [[('dc', 'ipa', 1)], [('dc', 'example', 1)]]),
        ('cn=R\\,W privilege', [[('cn', 'R,W privilege', 1)]]),
    ]
)
def test_str2dn2str(dnstring, expected):
    dn = str2dn(dnstring)
    assert dn == expected
    assert dn2str(dn) == dnstring
    assert dn_ctypes.str2dn(dnstring) == dn
    assert dn_ctypes.dn2str(dn) == dnstring


@pytest.mark.parametrize(
    'dnstring',
    [
        'cn',
        'cn=foo,',
        'cn=foo+bar',
    ]
)
def test_str2dn_errors(dnstring):
    with pytest.raises(DECODING_ERROR):
        str2dn(dnstring)
    with pytest.raises(dn_ctypes.DECODING_ERROR):
        dn_ctypes.str2dn(dnstring)


def test_dn2str_special():
    dnstring = 'cn=R\\2cW privilege'
    dnstring2 = 'cn=R\\,W privilege'
    expected = [[('cn', 'R,W privilege', 1)]]

    dn = str2dn(dnstring)
    assert dn == expected
    assert dn2str(dn) == dnstring2
    assert dn_ctypes.str2dn(dnstring) == dn
    assert dn_ctypes.dn2str(dn) == dnstring2



# Copyright (C) 2021  FreeIPA Project Contributors - see LICENSE file

from collections import namedtuple
from io import BytesIO
from lxml.etree import parse as myparse  # pylint: disable=no-name-in-module
import pytest
import textwrap
from unittest.mock import mock_open, patch

from ipaplatform.constants import constants
from ipaserver.install import dogtaginstance


class MyDogtagInstance(dogtaginstance.DogtagInstance):
    """Purpose is to avoid reading configuration files.

       The real DogtagInstance will open up the system store and
       try to determine the actual version of tomcat installed.
       Fake it instead.
    """
    def __init__(self, is_newer):
        self.service_user = constants.PKI_USER
        self.ajp_secret = None
        self.is_newer = is_newer

    def _is_newer_tomcat_version(self, default=None):
        return self.is_newer


def mock_etree_parse(data):
    """Convert a string into a file-like object to pass in the XML"""
    f = BytesIO(data.strip().encode('utf-8'))
    return myparse(f)


def mock_pkiuser_entity():
    """Return struct_passwd for mocked pkiuser"""
    StructPasswd = namedtuple(
        "StructPasswd",
        [
            "pw_name",
            "pw_passwd",
            "pw_uid",
            "pw_gid",
            "pw_gecos",
            "pw_dir",
            "pw_shell",
        ]
    )
    pkiuser_entity = StructPasswd(
        constants.PKI_USER,
        pw_passwd="x",
        pw_uid=-1,
        pw_gid=-1,
        pw_gecos="",
        pw_dir="/dev/null",
        pw_shell="/sbin/nologin",
    )
    return pkiuser_entity


# Format of test_data is:
#    (
#        is_newer_tomcat (boolean),
#        XML input,
#        expected secret attribute(s),
#        expected password(s),
#        rewrite of XML file expected (boolean),
#    )

test_data = (
    (
        #  Case 1: Upgrade requiredSecret to secret
        True,
        textwrap.dedent("""
            <?xml version="1.0" encoding="UTF-8"?>
            <Server port="1234" shutdown="SHUTDOWN">
              <Service name="Catalina">
                <Connector port="9000" protocol="AJP/1.3" redirectPort="443"
                 address="localhost" requiredSecret="testing_ajp_secret" />
              </Service>
            </Server>
        """),
        ('secret',),
        ('testing_ajp_secret',),
        True,
    ),
    (
        #  Case 2: One connector with secret, no update is needed
        True,
        textwrap.dedent("""
            <?xml version="1.0" encoding="UTF-8"?>
            <Server port="1234" shutdown="SHUTDOWN">
              <Service name="Catalina">
                <Connector port="9000" protocol="AJP/1.3" redirectPort="443"
                 address="localhost" secret="testing_ajp_secret" />
              </Service>
            </Server>
        """),
        ('secret',),
        ('testing_ajp_secret',),
        False,
    ),
    (
        #  Case 3: Two connectors, old secret attribute, different secrets
        True,
        textwrap.dedent("""
            <?xml version="1.0" encoding="UTF-8"?>
            <Server port="1234" shutdown="SHUTDOWN">
              <Service name="Catalina">
                <Connector port="9000" protocol="AJP/1.3" redirectPort="443"
                 address="localhost4" requiredSecret="testing_ajp_secret" />
                <Connector port="9000" protocol="AJP/1.3" redirectPort="443"
                 address="localhost6" requiredSecret="other_secret" />
              </Service>
            </Server>
        """),
        ('secret', 'secret'),
        ('testing_ajp_secret', 'testing_ajp_secret'),
        True,
    ),
    (
        #  Case 4: Two connectors, new secret attribute, same secrets
        True,
        textwrap.dedent("""
            <?xml version="1.0" encoding="UTF-8"?>
            <Server port="1234" shutdown="SHUTDOWN">
              <Service name="Catalina">
                <Connector port="9000" protocol="AJP/1.3" redirectPort="443"
                 address="localhost4" secret="testing_ajp_secret" />
                <Connector port="9000" protocol="AJP/1.3" redirectPort="443"
                 address="localhost6" secret="testing_ajp_secret" />
              </Service>
            </Server>
        """),
        ('secret', 'secret'),
        ('testing_ajp_secret', 'testing_ajp_secret'),
        False,
    ),
    (
        #  Case 5: Two connectors, no secrets
        True,
        textwrap.dedent("""
            <?xml version="1.0" encoding="UTF-8"?>
            <Server port="1234" shutdown="SHUTDOWN">
              <Service name="Catalina">
                <Connector port="9000" protocol="AJP/1.3" redirectPort="443"
                 address="localhost4" />
                <Connector port="9000" protocol="AJP/1.3" redirectPort="443"
                 address="localhost6" />
              </Service>
            </Server>
        """),
        ('secret', 'secret'),
        ('RANDOM', 'RANDOM'),
        True,
    ),
    (
        #  Case 6: Older tomcat, no update needed for requiredSecret
        False,
        textwrap.dedent("""
            <?xml version="1.0" encoding="UTF-8"?>
            <Server port="1234" shutdown="SHUTDOWN">
              <Service name="Catalina">
                <Connector port="9000" protocol="AJP/1.3" redirectPort="443"
                 address="localhost" requiredSecret="testing_ajp_secret" />
              </Service>
            </Server>
        """),
        ('requiredSecret',),
        ('testing_ajp_secret',),
        False,
    ),
)


class TestAJPSecretUpgrade:
    @patch("ipaplatform.base.constants.pwd.getpwnam")
    @patch("ipaplatform.base.constants.os.chown")
    @patch("ipaserver.install.dogtaginstance.lxml.etree.parse")
    @pytest.mark.parametrize("test_data", test_data)
    def test_connecter(self, mock_parse, mock_chown, mock_getpwnam, test_data):
        is_newer, data, secret, expect, rewrite = test_data
        mock_chown.return_value = None
        mock_parse.return_value = mock_etree_parse(data)
        mock_getpwnam.return_value = mock_pkiuser_entity()

        dogtag = MyDogtagInstance(is_newer)
        with patch('ipaserver.install.dogtaginstance.open', mock_open()) \
                as mocked_file:
            dogtag.secure_ajp_connector()
            if rewrite:
                newdata = mocked_file().write.call_args.args
                f = BytesIO(newdata[0])
                server_xml = myparse(f)
                doc = server_xml.getroot()
                connectors = doc.xpath('//Connector[@protocol="AJP/1.3"]')
                assert len(connectors) == len(secret)

                i = 0
                for connector in connectors:
                    if expect[i] == 'RANDOM':
                        assert connector.attrib[secret[i]]
                    else:
                        assert connector.attrib[secret[i]] == expect[i]
                    i += 1
            else:
                assert mocked_file().write.call_args is None



# Authors:
#   Petr Vobornik <pvoborni@redhat.com>
#
# Copyright (C) 2013  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

ENTITY = 'group'
DEFAULT_FACET = 'member_user'

PKEY = 'itest-group'
DATA = {
    'pkey': PKEY,
    'add': [
        ('textbox', 'cn', PKEY),
        ('textarea', 'description', 'test-group desc'),
        ('radio', 'type', 'nonposix'),
    ],
    'mod': [
        ('textarea', 'description', 'test-group desc modified'),
    ],
}

PKEY2 = 'itest-group2'
DATA2 = {
    'pkey': PKEY2,
    'add': [
        ('textbox', 'cn', PKEY2),
        ('textarea', 'description', 'test-group2 desc'),
    ]
}

PKEY3 = 'itest-group3'
DATA3 = {
    'pkey': PKEY3,
    'add': [
        ('textbox', 'cn', PKEY3),
        ('textarea', 'description', 'test-group3 desc'),
    ]
}

PKEY4 = 'itest-group4'
DATA4 = {
    'pkey': PKEY4,
    'add': [
        ('textbox', 'cn', PKEY4),
        ('textarea', 'description', 'test-group4 desc'),
    ]
}

PKEY5 = 'itest-group5'
DATA5 = {
    'pkey': PKEY5,
    'add': [
        ('textbox', 'cn', PKEY5),
        ('textarea', 'description', 'test-group5 desc'),
    ]
}

PKEY6 = 'itest-group6'
DATA6 = {
    'pkey': PKEY6,
    'add': [
        ('textbox', 'cn', PKEY6),
        ('textarea', 'description', 'test-group6 desc'),
        ('textbox', 'gidnumber', '77777'),
    ]
}

PKEY7 = ''
DATA7 = {
    'pkey': PKEY7,
    'add': [
        ('textbox', 'cn', PKEY7),
        ('textarea', 'description', 'Empty Group name'),
    ]
}

PKEY8 = ';test-gr@up'
DATA8 = {
    'pkey': PKEY8,
    'add': [
        ('textbox', 'cn', PKEY8),
        ('textarea', 'description', 'Invalid Group name'),
    ]
}

PKEY9 = 'itest-group9'
DATA9 = {
    'pkey': PKEY9,
    'add': [
        ('textbox', 'cn', PKEY9),
        ('textarea', 'description', 'test-group9 desc'),
        ('radio', 'type', 'nonposix'),
    ]
}

PKEY10 = 'itest-group10'
DATA10 = {
    'pkey': PKEY10,
    'add': [
        ('textbox', 'cn', PKEY10),
        ('textarea', 'description', 'test-group10 desc'),
        ('radio', 'type', 'nonposix'),
    ]
}

PKEY_SPECIAL_CHAR_GROUP = 'itest...group_-$'
DATA_SPECIAL_CHAR_GROUP = {
    'pkey': PKEY_SPECIAL_CHAR_GROUP,
    'add': [
        ('textbox', 'cn', PKEY_SPECIAL_CHAR_GROUP),
        ('textarea', 'description', 'special characters group desc'),
    ]
}



#
# Copyright (C) 2019  FreeIPA Contributors see COPYING for license
#

"""This module provides tests for SMB-related features like
   configuring Samba file server and mounting SMB file system
"""

from __future__ import absolute_import

from functools import partial
import textwrap
import re
import os

import pytest

from ipatests.test_integration.base import IntegrationTest
from ipatests.pytest_ipa.integration import tasks
from ipaplatform.osinfo import osinfo
from ipaplatform.paths import paths
from ipatests.pytest_ipa.integration import skip_if_fips


def wait_smbd_functional(host):
    """Wait smbd is functional after (re)start

    After start of smbd there is a 2-3 seconds delay before daemon is
    fully functional and clients can successfuly mount a share.
    The ping command effectively blocks until the daemon is ready.
    """
    host.run_command(['smbcontrol', 'smbd', 'ping'])


class TestSMB(IntegrationTest):
    topology = 'star'
    num_clients = 2
    num_ad_domains = 1

    ipa_user1 = 'user1'
    ipa_user1_password = 'SecretUser1'
    ipa_user2 = 'user2'
    ipa_user2_password = 'SecretUser2'
    ad_user_login = 'testuser'
    ad_user_password = 'Secret123'
    ipa_test_group = 'ipa_testgroup'
    ad_test_group = 'testgroup'

    @classmethod
    def install(cls, mh):
        if cls.domain_level is not None:
            domain_level = cls.domain_level
        else:
            domain_level = cls.master.config.domain_level
        tasks.install_topo(cls.topology,
                           cls.master, cls.replicas,
                           cls.clients, domain_level,
                           clients_extra_args=('--mkhomedir',))

        cls.ad = cls.ads[0]
        cls.smbserver = cls.clients[0]
        cls.smbclient = cls.clients[1]
        cls.ad_user = '{}@{}'.format(cls.ad_user_login, cls.ad.domain.name)

        tasks.install_adtrust(cls.master)
        tasks.configure_dns_for_trust(cls.master, cls.ad)
        tasks.configure_windows_dns_for_trust(cls.ad, cls.master)
        tasks.establish_trust_with_ad(cls.master, cls.ad.domain.name,
                                      extra_args=['--two-way=true'])

        tasks.create_active_user(cls.master, cls.ipa_user1,
                                 password=cls.ipa_user1_password)
        tasks.create_active_user(cls.master, cls.ipa_user2,
                                 password=cls.ipa_user2_password)
        # Trigger creation of home directories on the SMB server
        for user in [cls.ipa_user1, cls.ipa_user2, cls.ad_user]:
            tasks.run_command_as_user(cls.smbserver, user, ['stat', '.'])

    @pytest.fixture
    def samba_share_public(self):
        """Setup share outside /home on samba server."""
        share_name = 'shared'
        share_path = '/srv/samba_shared'
        smbserver = self.smbserver

        smbserver.run_command(['mkdir', share_path])
        smbserver.run_command(['chmod', '777', share_path])
        # apply selinux context only if selinux is enabled
        if tasks.is_selinux_enabled(smbserver):
            smbserver.run_command(['chcon', '-t', 'samba_share_t', share_path])
        with tasks.FileBackup(smbserver, paths.SMB_CONF):
            smb_conf = smbserver.get_file_contents(
                paths.SMB_CONF, encoding='utf-8')
            smb_conf += textwrap.dedent('''
            [{name}]
                path = {path}
                writable = yes
                browsable=yes
            '''.format(name=share_name, path=share_path))
            smbserver.put_file_contents(paths.SMB_CONF, smb_conf)
            smbserver.run_command(['systemctl', 'restart', 'smb'])
            wait_smbd_functional(smbserver)
            yield {
                'name': share_name,
                'server_path': share_path,
                'unc': '//{}/{}'.format(smbserver.hostname, share_name)
            }
        smbserver.run_command(['systemctl', 'restart', 'smb'])
        wait_smbd_functional(smbserver)
        smbserver.run_command(['rmdir', share_path])

    def mount_smb_share(self, user, password, share, mountpoint):
        tasks.kdestroy_all(self.smbclient)
        tasks.kinit_as_user(self.smbclient, user, password)
        self.smbclient.run_command(['mkdir', '-p', mountpoint])
        self.smbclient.run_command([
            'mount', '-t', 'cifs', share['unc'], mountpoint,
            '-o', 'sec=krb5i,multiuser'
        ])
        tasks.kdestroy_all(self.smbclient)

    def smb_sanity_check(self, user, client_mountpoint, share):
        test_dir = 'testdir_{}'.format(user)
        test_file = 'testfile_{}'.format(user)
        test_file_path = '{}/{}'.format(test_dir, test_file)
        test_string = 'Hello, world!'

        run_smb_client = partial(tasks.run_command_as_user, self.smbclient,
                                 user, cwd=client_mountpoint)
        run_smb_server = partial(self.smbserver.run_command,
                                 cwd=share['server_path'])

        try:
            # check creation of directory from client side
            run_smb_client(['mkdir', test_dir])
            # check dir properties at client side
            res = run_smb_client(['stat', '-c', '%n %U %G', test_dir])
            assert res.stdout_text == '{0} {1} {1}\n'.format(test_dir, user)
            # check dir properties at server side
            res = run_smb_server(['stat', '-c', '%n %U %G', test_dir])
            assert res.stdout_text == '{0} {1} {1}\n'.format(test_dir, user)

            # check creation of file from client side
            run_smb_client('printf "{}" > {}'.format(
                test_string, test_file_path))
            # check file is listed at client side
            res = run_smb_client(['ls', test_dir])
            assert res.stdout_text == test_file + '\n'
            # check file is listed at server side
            res = run_smb_server(['ls', test_dir])
            assert res.stdout_text == test_file + '\n'
            # check file properties at server side
            res = run_smb_server(['stat', '-c', '%n %s %U %G', test_file_path])
            assert res.stdout_text == '{0} {1} {2} {2}\n'.format(
                test_file_path, len(test_string), user)
            # check file properties at client side
            res = run_smb_client(['stat', '-c', '%n %s %U %G', test_file_path])
            assert res.stdout_text == '{0} {1} {2} {2}\n'.format(
                test_file_path, len(test_string), user)
            # check file contents at client side
            res = run_smb_client(['cat', test_file_path])
            assert res.stdout_text == test_string
            # check file contents at server side
            file_contents_at_server = self.smbserver.get_file_contents(
                '{}/{}'.format(share['server_path'], test_file_path),
                encoding='utf-8')
            assert file_contents_at_server == test_string

            # Detect whether smbclient uses -k or --use-kerberos=required
            # https://pagure.io/freeipa/issue/8926
            # then check access using smbclient.
            res = run_smb_client(
                [
                    "smbclient",
                    "-h",
                ], raiseonerr=False
            )
            if "[-k|--kerberos]" in res.stderr_text:
                smbclient_krb5_knob = "-k"
            else:
                smbclient_krb5_knob = "--use-kerberos=desired"
            res = run_smb_client(
                [
                    "smbclient",
                    smbclient_krb5_knob,
                    share["unc"],
                    "-c",
                    "dir",
                ]
            )
            assert test_dir in res.stdout_text

            # check file and dir removal from client side
            run_smb_client(['rm', test_file_path])
            run_smb_client(['rmdir', test_dir])
            # check dir does not exist at client side
            res = run_smb_client(['stat', test_dir], raiseonerr=False)
            assert res.returncode == 1
            assert 'No such file or directory' in res.stderr_text
            # check dir does not exist at server side
            res = run_smb_server(['stat', test_dir], raiseonerr=False)
            assert res.returncode == 1
            assert 'No such file or directory' in res.stderr_text
        finally:
            run_smb_server(['rm', '-rf', test_dir], raiseonerr=False)

    def smb_installation_check(self, result):
        domain_regexp_tpl = r'''
            Domain\ name:\s*{domain}\n
            \s*NetBIOS\ name:\s*{netbios}\n
            \s*SID:\s*S-1-5-21-\d+-\d+-\d+\n
            \s+ID\ range:\s*\d+\s*-\s*\d+
        '''
        # pylint: disable=no-member
        ipa_regexp = domain_regexp_tpl.format(
            domain=re.escape(self.master.domain.name),
            netbios=self.master.netbios)
        ad_regexp = domain_regexp_tpl.format(
            domain=re.escape(self.ad.domain.name), netbios=self.ad.netbios)
        # pylint: enable=no-member
        output_regexp = r'''
            Discovered\ domains.*
            {}
            .*
            {}
            .*
            Samba.+configured.+check.+/etc/samba/smb\.conf
        '''.format(ipa_regexp, ad_regexp)
        assert re.search(output_regexp, result.stdout_text,
                         re.VERBOSE | re.DOTALL)

    def cleanup_mount(self, mountpoint):
        self.smbclient.run_command(['umount', mountpoint], raiseonerr=False)
        self.smbclient.run_command(['rmdir', mountpoint], raiseonerr=False)

    def test_samba_uninstallation_without_installation(self):
        res = self.smbserver.run_command(
            ['ipa-client-samba', '--uninstall', '-U'])
        assert res.stdout_text == 'Samba domain member is not configured yet\n'

    def test_install_samba(self):
        samba_install_result = self.smbserver.run_command(
            ['ipa-client-samba', '-U'])
        # smb and winbind are expected to be not running
        for service in ['smb', 'winbind']:
            result = self.smbserver.run_command(
                ['systemctl', 'status', service], raiseonerr=False)
            assert result.returncode == 3
        self.smbserver.run_command([
            'systemctl', 'enable', '--now', 'smb', 'winbind'
        ])
        wait_smbd_functional(self.smbserver)
        # check that smb and winbind started successfully
        for service in ['smb', 'winbind']:
            self.smbserver.run_command(['systemctl', 'status', service])
        # print status for debugging purposes
        self.smbserver.run_command(['smbstatus'])
        # checks postponed till the end of method to be sure services are
        # started - this way we prevent other tests from failing
        self.smb_installation_check(samba_install_result)

    def test_authentication_with_smb_cifs_principal_alias(self):
        """Test that we can auth as NetBIOS alias

        cifs/... principal on SMB server side has NetBIOS name of the SMB
        server as its alias. Test that we can actually initialize credentials
        using this alias. We don't need to use it anywhere in Samba, just
        verify that alias works.
        Test for  https://pagure.io/freeipa/issue/8291"""
        netbiosname = self.smbserver.hostname.split('.')[0].upper() + '$'
        copier = tasks.KerberosKeyCopier(self.smbserver)

        principal = 'cifs/{hostname}@{realm}'.format(
            hostname=self.smbserver.hostname, realm=copier.realm)
        alias = '{netbiosname}@{realm}'.format(
            netbiosname=netbiosname, realm=copier.realm)
        replacement = {principal: alias}
        tmpname = tasks.create_temp_file(self.smbserver, create_file=False)
        try:
            copier.copy_keys(paths.SAMBA_KEYTAB, tmpname, principal=principal,
                             replacement=replacement)
            self.smbserver.run_command(['kinit', '-kt', tmpname, netbiosname])
        finally:
            self.smbserver.run_command(['rm', '-f', tmpname])

    def test_samba_service_listed(self):
        """Check samba service is listed.

        Regression test for https://bugzilla.redhat.com/show_bug.cgi?id=1731433
        """
        service_name = 'cifs/{}@{}'.format(
            self.smbserver.hostname, self.smbserver.domain.name.upper())
        tasks.kinit_admin(self.master)
        res = self.master.run_command(
            ['ipa', 'service-show', '--raw', service_name])
        expected_output = 'krbprincipalname: {}\n'.format(service_name)
        assert expected_output in res.stdout_text
        res = self.master.run_command(
            ['ipa', 'service-find', '--raw', service_name])
        assert expected_output in res.stdout_text

    def check_smb_access_at_ipa_client(self, user, password, samba_share):
        mount_point = '/mnt/smb'

        self.mount_smb_share(user, password, samba_share, mount_point)
        try:
            tasks.run_command_as_user(self.smbclient, user, ['kdestroy', '-A'])
            tasks.run_command_as_user(self.smbclient, user, ['kinit', user],
                                      stdin_text=password + '\n')
            self.smb_sanity_check(user, mount_point, samba_share)
        finally:
            self.cleanup_mount(mount_point)

    def test_smb_access_for_ipa_user_at_ipa_client(self):
        samba_share = {
            'name': 'homes',
            'server_path': '/home/{}'.format(self.ipa_user1),
            'unc': '//{}/homes'.format(self.smbserver.hostname)
        }
        self.check_smb_access_at_ipa_client(
            self.ipa_user1, self.ipa_user1_password, samba_share)

    def test_smb_access_for_ad_user_at_ipa_client(self):
        samba_share = {
            'name': 'homes',
            'server_path': '/home/{}/{}'.format(self.ad.domain.name,
                                                self.ad_user_login),
            'unc': '//{}/homes'.format(self.smbserver.hostname)
        }
        self.check_smb_access_at_ipa_client(
            self.ad_user, self.ad_user_password, samba_share)

    def test_smb_mount_and_access_by_different_users(self, samba_share_public):
        user1 = self.ipa_user1
        password1 = self.ipa_user1_password
        user2 = self.ipa_user2
        password2 = self.ipa_user2_password
        mount_point = '/mnt/smb'

        try:
            self.mount_smb_share(user1, password1, samba_share_public,
                                 mount_point)
            tasks.run_command_as_user(self.smbclient, user2,
                                      ['kdestroy', '-A'])
            tasks.run_command_as_user(self.smbclient, user2, ['kinit', user2],
                                      stdin_text=password2 + '\n')
            self.smb_sanity_check(user2, mount_point, samba_share_public)
        finally:
            self.cleanup_mount(mount_point)

    @pytest.mark.skipif(
        osinfo.id == 'fedora' and osinfo.version_number <= (31,),
        reason='Test requires krb 1.18')
    def test_smb_service_s4u2self(self):
        """Test S4U2Self operation by IPA service
           against both AD and IPA users
        """
        script = textwrap.dedent("""export KRB5_TRACE=/dev/stderr
        kdestroy -A
        kinit -kt /etc/samba/samba.keytab {principal}
        klist -f
        {print_pac} -k /etc/samba/samba.keytab -E impersonate {user_princ}
        klist -f
        """)
        principal = 'cifs/{hostname}'.format(
            hostname=self.smbserver.hostname)
        # Copy ipa-print-pac to SMB server
        # We can do so because Samba and GSSAPI libraries
        # are present there
        print_pac = self.master.get_file_contents(
            os.path.join(paths.LIBEXEC_IPA_DIR, "ipa-print-pac"))
        result = self.smbserver.run_command(['mktemp'])
        tmpname = result.stdout_text.strip()
        self.smbserver.put_file_contents(tmpname, print_pac)
        self.smbserver.run_command(['chmod', 'a+x', tmpname])
        for user in (self.ad_user, self.ipa_user1,):
            shell_script = script.format(principal=principal,
                                         user_princ=user,
                                         print_pac=tmpname)
            self.smbserver.run_command(['/bin/bash', '-s', '-e'],
                                       stdin_text=shell_script)
        self.smbserver.run_command(['rm', '-f', tmpname])
        tasks.kdestroy_all(self.smbserver)

    def test_smb_mount_fails_without_kerberos_ticket(self, samba_share_public):
        mountpoint = '/mnt/smb'
        try:
            tasks.kdestroy_all(self.smbclient)
            self.smbclient.run_command(['mkdir', '-p', mountpoint])
            res = self.smbclient.run_command([
                'mount', '-t', 'cifs', samba_share_public['unc'], mountpoint,
                '-o', 'sec=krb5i,multiuser'
            ], raiseonerr=False)
            assert res.returncode == 32
        finally:
            self.cleanup_mount(mountpoint)

    def check_repeated_smb_mount(self, options):
        mountpoint = '/mnt/smb'
        unc = '//{}/homes'.format(self.smbserver.hostname)
        test_file = 'ntlm_test'
        test_file_server_path = '/home/{}/{}'.format(self.ipa_user1, test_file)
        test_file_client_path = '{}/{}'.format(mountpoint, test_file)

        self.smbclient.run_command(['mkdir', '-p', mountpoint])
        self.smbserver.put_file_contents(test_file_server_path, '')
        try:
            for i in [1, 2]:
                res = self.smbclient.run_command([
                    'mount', '-t', 'cifs', unc, mountpoint, '-o', options],
                    raiseonerr=False)
                assert res.returncode == 0, (
                    'Mount failed at iteration {}. Output: {}'
                    .format(i, res.stdout_text + res.stderr_text))
                assert self.smbclient.transport.file_exists(
                    test_file_client_path)
                self.smbclient.run_command(['umount', mountpoint])
        finally:
            self.cleanup_mount(mountpoint)
            self.smbserver.run_command(['rm', '-f', test_file_server_path])

    @skip_if_fips()
    def test_ntlm_authentication_with_auto_domain(self):
        """Repeatedly try to authenticate with username and password with
        automatic domain discovery.

        This is a regression test for https://pagure.io/freeipa/issue/8636
        """
        tasks.kdestroy_all(self.smbclient)

        mount_options = 'user={user},pass={password},domainauto'.format(
            user=self.ipa_user1,
            password=self.ipa_user1_password
        )

        self.check_repeated_smb_mount(mount_options)

    @skip_if_fips()
    def test_ntlm_authentication_with_upn_with_lowercase_domain(self):
        tasks.kdestroy_all(self.smbclient)

        mount_options = 'user={user}@{domain},pass={password}'.format(
            user=self.ipa_user1,
            password=self.ipa_user1_password,
            domain=self.master.domain.name.lower()
        )
        self.check_repeated_smb_mount(mount_options)

    @skip_if_fips()
    def test_ntlm_authentication_with_upn_with_uppercase_domain(self):
        tasks.kdestroy_all(self.smbclient)

        mount_options = 'user={user}@{domain},pass={password}'.format(
            user=self.ipa_user1,
            password=self.ipa_user1_password,
            domain=self.master.domain.name.upper()
        )
        self.check_repeated_smb_mount(mount_options)

    def test_uninstall_samba(self):
        self.smbserver.run_command(['ipa-client-samba', '--uninstall', '-U'])
        res = self.smbserver.run_command(
            ['systemctl', 'status', 'winbind'], raiseonerr=False)
        assert res.returncode == 3
        res = self.smbserver.run_command(
            ['systemctl', 'status', 'smb'], raiseonerr=False)
        assert res.returncode == 3

    def test_repeated_uninstall_samba(self):
        """Test samba uninstallation after successful uninstallation.

        Test for bug https://pagure.io/freeipa/issue/8019.
        """
        self.smbserver.run_command(['ipa-client-samba', '--uninstall', '-U'])

    def test_samba_reinstall(self):
        """Test samba can be reinstalled.

        Test installation after uninstallation and do some sanity checks.
        Test for bug https://pagure.io/freeipa/issue/8021
        """
        self.test_install_samba()
        self.test_smb_access_for_ipa_user_at_ipa_client()

    def test_cleanup(self):
        tasks.unconfigure_windows_dns_for_trust(self.ad, self.master)



#
# ipachangeconf - configuration file manipulation classes and functions
# partially based on authconfig code
# Copyright (c) 1999-2007 Red Hat, Inc.
# Author: Simo Sorce <ssorce@redhat.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import warnings
from ipapython.ipachangeconf import IPAChangeConf as realIPAChangeConf


class IPAChangeConf(realIPAChangeConf):
    """Advertise the old name"""

    def __init__(self, name):
        """something"""
        warnings.warn(
            "Use 'ipapython.ipachangeconf.IPAChangeConfg'",
            DeprecationWarning,
            stacklevel=2
        )
        super(IPAChangeConf, self).__init__(name)



#
# Copyright (C) 2019  FreeIPA Contributors see COPYING for license
#
"""Tests for member manager feature
"""
from ipatests.test_integration.base import IntegrationTest
from ipatests.pytest_ipa.integration import tasks


PASSWORD = "DummyPassword123"
# direct member manager
USER_MM = "mmuser"
# indirect member manager through group membership
USER_INDIRECT = "indirect_mmuser"
GROUP_INDIRECT = "group_indirect"

USER1 = "testuser1"
USER2 = "testuser2"
GROUP1 = "testgroup1"
GROUP2 = "testgroup2"
HOSTGROUP1 = "testhostgroup1"


class TestMemberManager(IntegrationTest):
    """Tests for member manager feature for groups and hostgroups
    """
    topology = "line"

    @classmethod
    def install(cls, mh):
        super(TestMemberManager, cls).install(mh)
        master = cls.master

        tasks.create_active_user(master, USER_MM, PASSWORD)
        tasks.create_active_user(master, USER_INDIRECT, PASSWORD)
        tasks.create_active_user(master, USER1, PASSWORD)

        tasks.kinit_admin(master)
        tasks.group_add(master, GROUP_INDIRECT)
        master.run_command([
            'ipa', 'group-add-member', GROUP_INDIRECT, '--users', USER_INDIRECT
        ])

        tasks.user_add(master, USER2)
        tasks.group_add(master, GROUP1)
        tasks.group_add(master, GROUP2)
        master.run_command(['ipa', 'hostgroup-add', HOSTGROUP1])

        # make mmuser a member manager for group and hostgroup
        master.run_command([
            'ipa', 'group-add-member-manager', GROUP1,
            '--users', USER_MM
        ])
        master.run_command([
            'ipa', 'hostgroup-add-member-manager', HOSTGROUP1,
            '--users', USER_MM
        ])
        # make indirect group member manager for group and hostgroup
        master.run_command([
            'ipa', 'group-add-member-manager', GROUP1,
            '--groups', GROUP_INDIRECT
        ])
        master.run_command([
            'ipa', 'hostgroup-add-member-manager', HOSTGROUP1,
            '--groups', GROUP_INDIRECT
        ])
        tasks.kdestroy_all(master)

    def test_show_member_manager(self):
        master = self.master
        tasks.kinit_admin(master)

        result = master.run_command(['ipa', 'group-show', GROUP1])
        out = result.stdout_text
        assert f"Membership managed by groups: {GROUP_INDIRECT}" in out
        assert f"Membership managed by users: {USER_MM}" in out

        result = master.run_command(['ipa', 'hostgroup-show', HOSTGROUP1])
        out = result.stdout_text
        assert f"Membership managed by groups: {GROUP_INDIRECT}" in out
        assert f"Membership managed by users: {USER_MM}" in out

        tasks.kdestroy_all(master)

    def test_find_by_member_manager(self):
        master = self.master
        tasks.kinit_admin(master)

        result = master.run_command([
            'ipa', 'group-find', '--membermanager-users', USER_MM
        ])
        assert GROUP1 in result.stdout_text

        result = master.run_command([
            'ipa', 'group-find', '--membermanager-groups', GROUP_INDIRECT
        ])
        assert GROUP1 in result.stdout_text

        result = master.run_command(
            [
                'ipa', 'group-find', '--membermanager-users', USER1
            ],
            raiseonerr=False
        )
        assert result.returncode == 1
        assert "0 groups matched" in result.stdout_text

        result = master.run_command([
            'ipa', 'hostgroup-find', '--membermanager-users', USER_MM
        ])
        assert HOSTGROUP1 in result.stdout_text

        result = master.run_command([
            'ipa', 'hostgroup-find', '--membermanager-groups', GROUP_INDIRECT
        ])
        assert HOSTGROUP1 in result.stdout_text

        result = master.run_command(
            [
                'ipa', 'hostgroup-find', '--membermanager-users', USER1
            ],
            raiseonerr=False
        )
        assert result.returncode == 1
        assert "0 hostgroups matched" in result.stdout_text

    def test_group_member_manager_user(self):
        master = self.master
        # mmuser: add user1 to group
        tasks.kinit_as_user(master, USER_MM, PASSWORD)
        master.run_command([
            'ipa', 'group-add-member', GROUP1, '--users', USER1
        ])
        result = master.run_command(['ipa', 'group-show', GROUP1])
        assert USER1 in result.stdout_text

        # indirect: add user2 to group
        tasks.kinit_as_user(master, USER_INDIRECT, PASSWORD)
        master.run_command([
            'ipa', 'group-add-member', GROUP1, '--users', USER2
        ])
        # verify
        master.run_command(['ipa', 'group-show', GROUP1])
        result = master.run_command(['ipa', 'group-show', GROUP1])
        assert USER2 in result.stdout_text

    def test_group_member_manager_group(self):
        master = self.master
        # mmuser: add group2 to group
        tasks.kinit_as_user(master, USER_MM, PASSWORD)
        master.run_command([
            'ipa', 'group-add-member', GROUP1, '--groups', GROUP2
        ])
        result = master.run_command(['ipa', 'group-show', GROUP1])
        assert GROUP2 in result.stdout_text

    def test_group_member_manager_nopermission(self):
        master = self.master
        tasks.kinit_as_user(master, USER1, PASSWORD)
        result = master.run_command(
            [
                'ipa', 'group-add-member-manager', GROUP1, '--users', USER1
            ],
            raiseonerr=False
        )
        assert result.returncode != 0
        expected = (
            f"member user: {USER1}: Insufficient access: Insufficient "
            "'write' privilege to the 'memberManager' attribute of entry"
        )
        assert expected in result.stdout_text

    def test_hostgroup_member_manager_user(self):
        master = self.master
        # mmuser: add a host to host group
        tasks.kinit_as_user(master, USER_MM, PASSWORD)
        master.run_command([
            'ipa', 'hostgroup-add-member', HOSTGROUP1,
            '--hosts', master.hostname
        ])
        result = master.run_command(['ipa', 'hostgroup-show', HOSTGROUP1])
        assert master.hostname in result.stdout_text
        master.run_command([
            'ipa', 'hostgroup-remove-member', HOSTGROUP1,
            '--hosts', master.hostname
        ])
        result = master.run_command(['ipa', 'hostgroup-show', HOSTGROUP1])
        assert master.hostname not in result.stdout_text

        # indirect:
        tasks.kinit_as_user(master, USER_INDIRECT, PASSWORD)
        master.run_command([
            'ipa', 'hostgroup-add-member', HOSTGROUP1,
            '--hosts', master.hostname
        ])
        result = master.run_command(['ipa', 'hostgroup-show', HOSTGROUP1])
        assert master.hostname in result.stdout_text

    def test_hostgroup_member_manager_nopermission(self):
        master = self.master
        tasks.kinit_as_user(master, USER1, PASSWORD)
        result = master.run_command(
            [
                'ipa', 'hostgroup-add-member-manager', HOSTGROUP1,
                '--users', USER1
            ],
            raiseonerr=False
        )
        assert result.returncode != 0
        expected = (
            f"member user: {USER1}: Insufficient access: Insufficient "
            "'write' privilege to the 'memberManager' attribute of entry"
        )
        assert expected in result.stdout_text



# Authors: Simo Sorce <ssorce@redhat.com>
#          Alexander Bokovoy <abokovoy@redhat.com>
#          Martin Kosek <mkosek@redhat.com>
#          Tomas Babej <tbabej@redhat.com>
#
# Copyright (C) 2007-2014  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
This module contains default Red Hat OS family-specific implementations of
system tasks.
'''
from __future__ import print_function, absolute_import

import ctypes
import logging
import os
from pathlib import Path
import socket
import traceback
import errno
import urllib
import subprocess
import sys
import textwrap

from ctypes.util import find_library
from functools import total_ordering
from subprocess import CalledProcessError

from pyasn1.error import PyAsn1Error

from ipapython import directivesetter
from ipapython import ipautil
import ipapython.errors

from ipaplatform.constants import constants
from ipaplatform.paths import paths
from ipaplatform.redhat.authconfig import get_auth_tool
from ipaplatform.base.tasks import BaseTaskNamespace

logger = logging.getLogger(__name__)


# /etc/pkcs11/modules override
# base filename, module, list of disabled-in
# 'p11-kit-proxy' disables proxying of module, see man(5) pkcs11.conf
PKCS11_MODULES = [
    ('softhsm2', paths.LIBSOFTHSM2_SO, ['p11-kit-proxy']),
]


NM_IPA_CONF = textwrap.dedent("""
    # auto-generated by IPA installer
    [main]
    dns={dnsprocessing}

    [global-dns]
    searches={searches}

    [global-dns-domain-*]
    servers={servers}
""")


@total_ordering
class IPAVersion:
    _rpmvercmp_func = None

    @classmethod
    def _rpmvercmp(cls, a, b):
        """Lazy load and call librpm's rpmvercmp
        """
        rpmvercmp_func = cls._rpmvercmp_func
        if rpmvercmp_func is None:
            librpm = ctypes.CDLL(find_library('rpm'))
            rpmvercmp_func = librpm.rpmvercmp
            # int rpmvercmp(const char *a, const char *b)
            rpmvercmp_func.argtypes = [ctypes.c_char_p, ctypes.c_char_p]
            rpmvercmp_func.restype = ctypes.c_int
            cls._rpmvercmp_func = rpmvercmp_func
        return rpmvercmp_func(a, b)

    def __init__(self, version):
        self._version = version
        self._bytes = version.encode('utf-8')

    @property
    def version(self):
        return self._version

    def __eq__(self, other):
        if not isinstance(other, IPAVersion):
            return NotImplemented
        return self._rpmvercmp(self._bytes, other._bytes) == 0

    def __lt__(self, other):
        if not isinstance(other, IPAVersion):
            return NotImplemented
        return self._rpmvercmp(self._bytes, other._bytes) < 0

    def __hash__(self):
        return hash(self._version)


class RedHatTaskNamespace(BaseTaskNamespace):

    def restore_context(self, filepath, force=False):
        """Restore SELinux security context on the given filepath.

        SELinux equivalent is /path/to/restorecon <filepath>
        restorecon's return values are not reliable so we have to
        ignore them (BZ #739604).

        ipautil.run() will do the logging.
        """
        restorecon = paths.SBIN_RESTORECON
        if not self.is_selinux_enabled() or not os.path.exists(restorecon):
            return

        # Force reset of context to match file_context for customizable
        # files, and the default file context, changing the user, role,
        # range portion as well as the type.
        args = [restorecon]
        if force:
            args.append('-F')
        args.append(filepath)
        ipautil.run(args, raiseonerr=False)

    def is_selinux_enabled(self):
        """Check if SELinux is available and enabled
        """
        try:
            ipautil.run([paths.SELINUXENABLED])
        except ipautil.CalledProcessError:
            # selinuxenabled returns 1 if not enabled
            return False
        except OSError:
            # selinuxenabled binary not available
            return False
        else:
            return True

    def check_selinux_status(self, restorecon=paths.RESTORECON):
        """
        We don't have a specific package requirement for policycoreutils
        which provides restorecon. This is because we don't require
        SELinux on client installs. However if SELinux is enabled then
        this package is required.

        This function returns nothing but may raise a Runtime exception
        if SELinux is enabled but restorecon is not available.
        """
        if not self.is_selinux_enabled():
            return False

        if not os.path.exists(restorecon):
            raise RuntimeError('SELinux is enabled but %s does not exist.\n'
                               'Install the policycoreutils package and start '
                               'the installation again.' % restorecon)
        return True

    def check_ipv6_stack_enabled(self):
        """Checks whether IPv6 kernel module is loaded.

        Function checks if /proc/net/if_inet6 is present. If IPv6 stack is
        enabled, it exists and contains the interfaces configuration.

        :raises: RuntimeError when IPv6 stack is disabled
        """
        if not os.path.exists(paths.IF_INET6):
            raise RuntimeError(
                "IPv6 stack has to be enabled in the kernel and some "
                "interface has to have ::1 address assigned. Typically "
                "this is 'lo' interface. If you do not wish to use IPv6 "
                "globally, disable it on the specific interfaces in "
                "sysctl.conf except 'lo' interface.")

        try:
            localhost6 = ipautil.CheckedIPAddress('::1', allow_loopback=True)
            if localhost6.get_matching_interface() is None:
                raise ValueError("no interface for ::1 address found")
        except ValueError:
            raise RuntimeError(
                 "IPv6 stack is enabled in the kernel but there is no "
                 "interface that has ::1 address assigned. Add ::1 address "
                 "resolution to 'lo' interface. You might need to enable IPv6 "
                 "on the interface 'lo' in sysctl.conf.")

    def detect_container(self):
        """Check if running inside a container

        :returns: container runtime or None
        :rtype: str, None
        """
        try:
            output = subprocess.check_output(
                [paths.SYSTEMD_DETECT_VIRT, '--container'],
                stderr=subprocess.STDOUT
            )
        except subprocess.CalledProcessError as e:
            if e.returncode == 1:
                # No container runtime detected
                return None
            else:
                raise
        else:
            return output.decode('utf-8').strip()

    def restore_pre_ipa_client_configuration(self, fstore, statestore,
                                             was_sssd_installed,
                                             was_sssd_configured):

        auth_config = get_auth_tool()
        auth_config.unconfigure(
            fstore, statestore, was_sssd_installed, was_sssd_configured
        )

    def set_nisdomain(self, nisdomain):
        try:
            with open(paths.SYSCONF_NETWORK, 'r') as f:
                content = [
                    line for line in f
                    if not line.strip().upper().startswith('NISDOMAIN')
                ]
        except IOError:
            content = []

        content.append("NISDOMAIN={}\n".format(nisdomain))

        with open(paths.SYSCONF_NETWORK, 'w') as f:
            f.writelines(content)

    def modify_nsswitch_pam_stack(self, sssd, mkhomedir, statestore,
                                  sudo=True):
        auth_config = get_auth_tool()
        auth_config.configure(sssd, mkhomedir, statestore, sudo)

    def is_nosssd_supported(self):
        # The flag --no-sssd is not supported any more for rhel-based distros
        return False

    def backup_auth_configuration(self, path):
        auth_config = get_auth_tool()
        auth_config.backup(path)

    def restore_auth_configuration(self, path):
        auth_config = get_auth_tool()
        auth_config.restore(path)

    def migrate_auth_configuration(self, statestore):
        """
        Migrate the pam stack configuration from authconfig to an authselect
        profile.
        """
        # Check if mkhomedir was enabled during installation
        mkhomedir = statestore.get_state('authconfig', 'mkhomedir')

        # Force authselect 'sssd' profile
        authselect_cmd = [paths.AUTHSELECT, "select", "sssd", "with-sudo"]
        if mkhomedir:
            authselect_cmd.append("with-mkhomedir")
        authselect_cmd.append("--force")
        ipautil.run(authselect_cmd)

        # Remove all remaining keys from the authconfig module
        for conf in ('ldap', 'krb5', 'sssd', 'sssdauth', 'mkhomedir'):
            statestore.restore_state('authconfig', conf)

        # Create new authselect module in the statestore
        statestore.backup_state('authselect', 'profile', 'sssd')
        statestore.backup_state(
            'authselect', 'features_list', '')
        statestore.backup_state('authselect', 'mkhomedir', bool(mkhomedir))

    def reload_systemwide_ca_store(self):
        try:
            ipautil.run([paths.UPDATE_CA_TRUST])
        except CalledProcessError as e:
            logger.error(
                "Could not update systemwide CA trust database: %s", e)
            return False
        else:
            logger.info("Systemwide CA database updated.")
            return True

    def platform_insert_ca_certs(self, ca_certs):
        return any([
            self.write_p11kit_certs(paths.IPA_P11_KIT, ca_certs),
            self.remove_ca_certificates_bundle(
                paths.SYSTEMWIDE_IPA_CA_CRT
            ),
        ])

    def write_p11kit_certs(self, filename, ca_certs):
        # pylint: disable=ipa-forbidden-import
        from ipalib import x509  # FixMe: break import cycle
        from ipalib.errors import CertificateError
        # pylint: enable=ipa-forbidden-import

        path = Path(filename)
        try:
            f = open(path, 'w')
        except IOError:
            logger.error("Failed to open %s", path)
            raise

        with f:
            f.write("# This file was created by IPA. Do not edit.\n"
                    "\n")

            try:
                os.fchmod(f.fileno(), 0o644)
            except IOError:
                logger.error("Failed to set mode of %s", path)
                raise

            has_eku = set()
            for cert, nickname, trusted, _ext_key_usage in ca_certs:
                try:
                    subject = cert.subject_bytes
                    issuer = cert.issuer_bytes
                    serial_number = cert.serial_number_bytes
                    public_key_info = cert.public_key_info_bytes
                except (PyAsn1Error, ValueError, CertificateError):
                    logger.error(
                        "Failed to decode certificate \"%s\"", nickname)
                    raise

                label = urllib.parse.quote(nickname)
                subject = urllib.parse.quote(subject)
                issuer = urllib.parse.quote(issuer)
                serial_number = urllib.parse.quote(serial_number)
                public_key_info = urllib.parse.quote(public_key_info)

                obj = ("[p11-kit-object-v1]\n"
                       "class: certificate\n"
                       "certificate-type: x-509\n"
                       "certificate-category: authority\n"
                       "label: \"%(label)s\"\n"
                       "subject: \"%(subject)s\"\n"
                       "issuer: \"%(issuer)s\"\n"
                       "serial-number: \"%(serial_number)s\"\n"
                       "x-public-key-info: \"%(public_key_info)s\"\n" %
                       dict(label=label,
                            subject=subject,
                            issuer=issuer,
                            serial_number=serial_number,
                            public_key_info=public_key_info))
                if trusted is True:
                    obj += "trusted: true\n"
                elif trusted is False:
                    obj += "x-distrusted: true\n"
                obj += "{pem}\n\n".format(
                    pem=cert.public_bytes(x509.Encoding.PEM).decode('ascii'))

                f.write(obj)

                if (cert.extended_key_usage is not None and
                        public_key_info not in has_eku):
                    try:
                        ext_key_usage = cert.extended_key_usage_bytes
                    except PyAsn1Error:
                        logger.error(
                            "Failed to encode extended key usage for \"%s\"",
                            nickname)
                        raise
                    value = urllib.parse.quote(ext_key_usage)
                    obj = ("[p11-kit-object-v1]\n"
                           "class: x-certificate-extension\n"
                           "label: \"ExtendedKeyUsage for %(label)s\"\n"
                           "x-public-key-info: \"%(public_key_info)s\"\n"
                           "object-id: 2.5.29.37\n"
                           "value: \"%(value)s\"\n\n" %
                           dict(label=label,
                                public_key_info=public_key_info,
                                value=value))
                    f.write(obj)
                    has_eku.add(public_key_info)

        return True

    def platform_remove_ca_certs(self):
        return any([
            self.remove_ca_certificates_bundle(paths.IPA_P11_KIT),
            self.remove_ca_certificates_bundle(paths.SYSTEMWIDE_IPA_CA_CRT),
        ])

    def remove_ca_certificates_bundle(self, filename):
        path = Path(filename)
        if not path.is_file():
            return False

        try:
            path.unlink()
        except Exception:
            logger.error("Could not remove %s", path)
            raise

        return True

    def backup_hostname(self, fstore, statestore):
        filepath = paths.ETC_HOSTNAME
        if os.path.exists(filepath):
            fstore.backup_file(filepath)

        # store old hostname
        old_hostname = socket.gethostname()
        statestore.backup_state('network', 'hostname', old_hostname)

    def restore_hostname(self, fstore, statestore):
        old_hostname = statestore.restore_state('network', 'hostname')

        if old_hostname is not None:
            try:
                self.set_hostname(old_hostname)
            except ipautil.CalledProcessError as e:
                logger.debug("%s", traceback.format_exc())
                logger.error(
                    "Failed to restore this machine hostname to %s (%s).",
                    old_hostname, e
                )

        filepath = paths.ETC_HOSTNAME
        if fstore.has_file(filepath):
            fstore.restore_file(filepath)

    def set_selinux_booleans(self, required_settings, backup_func=None):
        def get_setsebool_args(changes):
            args = [paths.SETSEBOOL, "-P"]
            args.extend(["%s=%s" % update for update in changes.items()])

            return args

        if not self.is_selinux_enabled():
            return False

        updated_vars = {}
        failed_vars = {}
        for setting, state in required_settings.items():
            if state is None:
                continue
            try:
                result = ipautil.run(
                    [paths.GETSEBOOL, setting],
                    capture_output=True
                )
                original_state = result.output.split()[2]
                if backup_func is not None:
                    backup_func(setting, original_state)

                if original_state != state:
                    updated_vars[setting] = state
            except ipautil.CalledProcessError as e:
                logger.error("Cannot get SELinux boolean '%s': %s", setting, e)
                failed_vars[setting] = state

        if updated_vars:
            args = get_setsebool_args(updated_vars)
            try:
                ipautil.run(args)
            except ipautil.CalledProcessError:
                failed_vars.update(updated_vars)

        if failed_vars:
            raise ipapython.errors.SetseboolError(
                failed=failed_vars,
                command=' '.join(get_setsebool_args(failed_vars)))

        return True

    def parse_ipa_version(self, version):
        """
        :param version: textual version
        :return: object implementing proper __cmp__ method for version compare
        """
        return IPAVersion(version)

    def configure_httpd_service_ipa_conf(self):
        """Create systemd config for httpd service to work with IPA
        """
        if not os.path.exists(paths.SYSTEMD_SYSTEM_HTTPD_D_DIR):
            os.mkdir(paths.SYSTEMD_SYSTEM_HTTPD_D_DIR, 0o755)

        ipautil.copy_template_file(
            os.path.join(paths.USR_SHARE_IPA_DIR, 'ipa-httpd.conf.template'),
            paths.SYSTEMD_SYSTEM_HTTPD_IPA_CONF,
            dict(
                KDCPROXY_CONFIG=paths.KDCPROXY_CONFIG,
                IPA_HTTPD_KDCPROXY=paths.IPA_HTTPD_KDCPROXY,
                KRB5CC_HTTPD=paths.KRB5CC_HTTPD,
            )
        )

        os.chmod(paths.SYSTEMD_SYSTEM_HTTPD_IPA_CONF, 0o644)
        self.restore_context(paths.SYSTEMD_SYSTEM_HTTPD_IPA_CONF)
        self.systemd_daemon_reload()

    def systemd_daemon_reload(self):
        """Tell systemd to reload config files"""
        ipautil.run([paths.SYSTEMCTL, "--system", "daemon-reload"])

    def configure_http_gssproxy_conf(self, ipaapi_user):
        ipautil.copy_template_file(
            os.path.join(paths.USR_SHARE_IPA_DIR, 'gssproxy.conf.template'),
            paths.GSSPROXY_CONF,
            dict(
                HTTP_KEYTAB=paths.HTTP_KEYTAB,
                HTTPD_USER=constants.HTTPD_USER,
                IPAAPI_USER=ipaapi_user,
                SWEEPER_SOCKET=paths.IPA_CCACHE_SWEEPER_GSSPROXY_SOCK,
            )
        )

        os.chmod(paths.GSSPROXY_CONF, 0o600)
        self.restore_context(paths.GSSPROXY_CONF)

    def configure_httpd_wsgi_conf(self):
        """Configure WSGI for correct Python version (Fedora)

        See https://pagure.io/freeipa/issue/7394
        """
        conf = paths.HTTPD_IPA_WSGI_MODULES_CONF
        if sys.version_info.major == 2:
            wsgi_module = constants.MOD_WSGI_PYTHON2
        else:
            wsgi_module = constants.MOD_WSGI_PYTHON3

        if conf is None or wsgi_module is None:
            logger.info("Nothing to do for configure_httpd_wsgi_conf")
            return

        confdir = os.path.dirname(conf)
        if not os.path.isdir(confdir):
            os.makedirs(confdir)

        ipautil.copy_template_file(
            os.path.join(
                paths.USR_SHARE_IPA_DIR, 'ipa-httpd-wsgi.conf.template'
            ),
            conf,
            dict(WSGI_MODULE=wsgi_module)
        )

        os.chmod(conf, 0o644)
        self.restore_context(conf)

    def remove_httpd_service_ipa_conf(self):
        """Remove systemd config for httpd service of IPA"""
        try:
            os.unlink(paths.SYSTEMD_SYSTEM_HTTPD_IPA_CONF)
        except OSError as e:
            if e.errno == errno.ENOENT:
                logger.debug(
                    'Trying to remove %s but file does not exist',
                    paths.SYSTEMD_SYSTEM_HTTPD_IPA_CONF
                )
            else:
                logger.error(
                    'Error removing %s: %s',
                    paths.SYSTEMD_SYSTEM_HTTPD_IPA_CONF, e
                )
            return

        self.systemd_daemon_reload()

    def configure_httpd_protocol(self):
        # use default crypto policy for SSLProtocol
        directivesetter.set_directive(
            paths.HTTPD_SSL_CONF, 'SSLProtocol', None, False
        )

    def set_hostname(self, hostname):
        ipautil.run([paths.BIN_HOSTNAMECTL, 'set-hostname', hostname])

    def is_fips_enabled(self):
        """
        Checks whether this host is FIPS-enabled.

        Returns a boolean indicating if the host is FIPS-enabled, i.e. if the
        file /proc/sys/crypto/fips_enabled contains a non-0 value. Otherwise,
        or if the file /proc/sys/crypto/fips_enabled does not exist,
        the function returns False.
        """
        try:
            with open(paths.PROC_FIPS_ENABLED, 'r') as f:
                if f.read().strip() != '0':
                    return True
        except IOError:
            # Consider that the host is not fips-enabled if the file does not
            # exist
            pass
        return False

    def setup_httpd_logging(self):
        directivesetter.set_directive(paths.HTTPD_SSL_CONF,
                                      'ErrorLog',
                                      'logs/error_log', False)
        directivesetter.set_directive(paths.HTTPD_SSL_CONF,
                                      'TransferLog',
                                      'logs/access_log', False)

    def configure_dns_resolver(self, nameservers, searchdomains, *,
                               resolve1_enabled=False, fstore=None):
        """Configure global DNS resolver (e.g. /etc/resolv.conf)

        :param nameservers: list of IP addresses
        :param searchdomains: list of search domaons
        :param fstore: optional file store for backup
        """
        assert nameservers and isinstance(nameservers, list)
        assert searchdomains and isinstance(searchdomains, list)

        super().configure_dns_resolver(
            nameservers=nameservers,
            searchdomains=searchdomains,
            resolve1_enabled=resolve1_enabled,
            fstore=fstore
        )

        # break circular import
        from ipaplatform.services import knownservices

        if fstore is not None and not fstore.has_file(paths.RESOLV_CONF):
            fstore.backup_file(paths.RESOLV_CONF)

        nm = knownservices['NetworkManager']
        nm_enabled = nm.is_enabled()
        if nm_enabled:
            logger.debug(
                "Network Manager is enabled, write %s",
                paths.NETWORK_MANAGER_IPA_CONF
            )
            # write DNS override and reload network manager to have it create
            # a new resolv.conf. The file is prefixed with ``zzz`` to
            # make it the last file. Global dns options do not stack and last
            # man standing wins.
            if resolve1_enabled:
                # push DNS configuration to systemd-resolved
                dnsprocessing = "systemd-resolved"
            else:
                # update /etc/resolv.conf
                dnsprocessing = "default"

            cfg = NM_IPA_CONF.format(
                dnsprocessing=dnsprocessing,
                servers=','.join(nameservers),
                searches=','.join(searchdomains)
            )
            with open(paths.NETWORK_MANAGER_IPA_CONF, 'w') as f:
                os.fchmod(f.fileno(), 0o644)
                f.write(cfg)
            # reload NetworkManager
            nm.reload_or_restart()

        if not resolve1_enabled and not nm_enabled:
            # no NM running, no systemd-resolved detected
            # fall back to /etc/resolv.conf
            logger.debug(
                "Neither Network Manager nor systemd-resolved are enabled, "
                "write %s directly.", paths.RESOLV_CONF
            )
            cfg = [
                "# auto-generated by IPA installer",
                "search {}".format(' '.join(searchdomains)),
            ]
            for nameserver in nameservers:
                cfg.append("nameserver {}".format(nameserver))
            with open(paths.RESOLV_CONF, 'w') as f:
                f.write('\n'.join(cfg))

    def unconfigure_dns_resolver(self, fstore=None):
        """Unconfigure global DNS resolver (e.g. /etc/resolv.conf)

        :param fstore: optional file store for restore
        """
        super().unconfigure_dns_resolver(fstore=fstore)
        # break circular import
        from ipaplatform.services import knownservices

        nm = knownservices['NetworkManager']
        if os.path.isfile(paths.NETWORK_MANAGER_IPA_CONF):
            os.unlink(paths.NETWORK_MANAGER_IPA_CONF)
            if nm.is_enabled():
                nm.reload_or_restart()

    def configure_pkcs11_modules(self, fstore):
        """Disable global p11-kit configuration for NSS
        """
        filenames = []
        for name, module, disabled_in in PKCS11_MODULES:
            filename = os.path.join(
                paths.ETC_PKCS11_MODULES_DIR,
                "{}.module".format(name)
            )
            if os.path.isfile(filename):
                # Only back up if file is not yet backed up and it does not
                # look like a file that is generated by IPA.
                with open(filename) as f:
                    content = f.read()
                is_ipa_file = "IPA" in content
                if not is_ipa_file and not fstore.has_file(filename):
                    logger.debug("Backing up existing '%s'.", filename)
                    fstore.backup_file(filename)

            with open(filename, "w") as f:
                f.write("# created by IPA installer\n")
                f.write("module: {}\n".format(module))
                # see man(5) pkcs11.conf
                f.write("disable-in: {}\n".format(", ".join(disabled_in)))
                os.fchmod(f.fileno(), 0o644)
            self.restore_context(filename)
            logger.debug("Created PKCS#11 module config '%s'.", filename)
            filenames.append(filename)

        return filenames

    def restore_pkcs11_modules(self, fstore):
        """Restore global p11-kit configuration for NSS
        """
        filenames = []
        for name, _module, _disabled_in in PKCS11_MODULES:
            filename = os.path.join(
                paths.ETC_PKCS11_MODULES_DIR,
                "{}.module".format(name)
            )
            try:
                os.unlink(filename)
            except OSError:
                pass
            else:
                filenames.append(filename)

            if fstore.has_file(filename):
                fstore.restore_file(filename)

        return filenames

    def get_pkcs11_modules(self):
        """Return the list of module config files setup by IPA
        """
        return tuple(os.path.join(paths.ETC_PKCS11_MODULES_DIR,
                                  "{}.module".format(name))
                     for name, _module, _disabled in PKCS11_MODULES)

    def enable_sssd_sudo(self, _fstore):
        """sudo enablement is handled by authselect"""

    def enable_ldap_automount(self, statestore):
        """
        Point automount to ldap in nsswitch.conf.
        This function is for non-SSSD setups only.
        """
        super(RedHatTaskNamespace, self).enable_ldap_automount(statestore)

        authselect_cmd = [paths.AUTHSELECT, "enable-feature",
                          "with-custom-automount"]
        ipautil.run(authselect_cmd)

    def disable_ldap_automount(self, statestore):
        """Disable ldap-based automount"""
        super(RedHatTaskNamespace, self).disable_ldap_automount(statestore)

        authselect_cmd = [paths.AUTHSELECT, "disable-feature",
                          "with-custom-automount"]
        ipautil.run(authselect_cmd)

tasks = RedHatTaskNamespace()



#
# Copyright (C) 2020  FreeIPA Contributors see COPYING for license
#
"""Get host's FQDN
"""
import socket


def gethostfqdn():
    """Get the fully qualified domain name of current host from glibc

    This function may return an FQDN with up to MAXHOSTFQDNLEN characters
    (253). The effective hostname is still limited to MAXHOSTNAMELEN (64).

    :return: FQDN as str
    """
    hostname = socket.gethostname()
    # this call can never fail except for misconfigured nsswitch.conf
    # without nss-myhostname provider. The myhostname provider translates
    # gethostname() to local interfaces.
    gai = socket.getaddrinfo(
        hostname,
        None,  # service/port is irrelevant
        family=socket.AF_UNSPEC,  # IPv4 or IPv6
        type=socket.SOCK_DGRAM,  # optimization, TCP/RAW gives same result
        # include canonical name in first addrinfo struct
        # only use address family when at least one non-local interface
        # is configured with that address family
        flags=socket.AI_CANONNAME | socket.AI_ADDRCONFIG
    )
    # first addrinfo struct, fourth field is canonical name
    # getaddrinfo() either raises an exception or returns at least one entry
    return gai[0][3]



# Authors:
#   Ana Krivokapic <akrivoka@redhat.com>
#
# Copyright (C) 2013  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Test adding/removing external members (trusted domain objects) to IPA groups.
These tests are skipped if trust is not established.
"""

from ipalib import api
from ipapython.dn import DN
from ipatests.test_xmlrpc import objectclasses
from ipatests.test_xmlrpc.xmlrpc_test import (Declarative, fuzzy_uuid,
                                              fuzzy_user_or_group_sid)
import pytest

group_name = u'external_group'
group_desc = u'Test external group'
group_dn = DN(('cn', group_name), api.env.container_group, api.env.basedn)


def get_trusted_group_name():
    trusts = api.Command['trust_find']()
    if trusts['count'] == 0:
        return None

    ad_netbios = trusts['result'][0]['ipantflatname']
    return r'%s\Domain Admins' % ad_netbios


@pytest.mark.tier1
class test_external_members(Declarative):
    @pytest.fixture(autouse=True, scope="class")
    def ext_member_setup(self, declarative_setup):
        if not api.Backend.rpcclient.isconnected():
            api.Backend.rpcclient.connect()

        trusts = api.Command['trust_find']()
        if trusts['count'] == 0:
            pytest.skip('Trust is not established')

    cleanup_commands = [
        ('group_del', [group_name], {}),
    ]

    tests = [
        dict(
            desc='Create external group "%s"' % group_name,
            command=(
                'group_add', [group_name], dict(description=group_desc, external=True)
                ),
            expected=dict(
                value=group_name,
                summary=u'Added group "%s"' % group_name,
                result=dict(
                    cn=[group_name],
                    description=[group_desc],
                    objectclass=objectclasses.externalgroup,
                    ipauniqueid=[fuzzy_uuid],
                    dn=group_dn,
                ),
            ),
        ),
        dict(
            desc='Add external member "%s" to group "%s"' % (get_trusted_group_name(), group_name),
            command=(
                'group_add_member', [group_name], dict(ipaexternalmember=get_trusted_group_name())
            ),
            expected=dict(
                completed=1,
                failed=dict(
                    member=dict(
                        group=tuple(),
                        user=tuple(),
                    ),
                ),
                result=dict(
                        dn=group_dn,
                        ipaexternalmember=[fuzzy_user_or_group_sid],
                        cn=[group_name],
                        description=[group_desc],
                ),
            ),
        ),
        dict(
            desc='Try to add duplicate external member "%s" to group "%s"' % (get_trusted_group_name(), group_name),
            command=(
                'group_add_member', [group_name], dict(ipaexternalmember=get_trusted_group_name())
            ),
            expected=dict(
                completed=0,
                failed=dict(
                    member=dict(
                        group=[(fuzzy_user_or_group_sid, u'This entry is already a member')],
                        user=tuple(),
                    ),
                ),
                result=dict(
                        dn=group_dn,
                        ipaexternalmember=[fuzzy_user_or_group_sid],
                        cn=[group_name],
                        description=[group_desc],
                ),
            ),
        ),
        dict(
            desc='Remove external member "%s" from group "%s"' % (get_trusted_group_name(), group_name),
            command=(
                'group_remove_member', [group_name], dict(ipaexternalmember=get_trusted_group_name())
            ),
            expected=dict(
                completed=1,
                failed=dict(
                    member=dict(
                        group=tuple(),
                        user=tuple(),
                    ),
                ),
                result=dict(
                    dn=group_dn,
                    cn=[group_name],
                    ipaexternalmember=[],
                    description=[group_desc],
                ),
            ),
        ),
        dict(
            desc='Try to remove external entry "%s" which is not a member of group "%s" from group "%s"' % (get_trusted_group_name(), group_name, group_name),
            command=(
                'group_remove_member', [group_name], dict(ipaexternalmember=get_trusted_group_name())
            ),
            expected=dict(
                completed=0,
                failed=dict(
                    member=dict(
                        group=[(fuzzy_user_or_group_sid, u'This entry is not a member')],
                        user=tuple(),
                    ),
                ),
                result=dict(
                        dn=group_dn,
                        cn=[group_name],
                        description=[group_desc],
                ),
            ),
        ),
    ]
